"""Tests for AbstractFileSystem implementation.

These tests verify that AbstractFileSystem can be subclassed to provide
a virtual filesystem that Sandbox code can interact with via Path methods.
"""

from pathlib import PurePosixPath

import pytest
from inline_snapshot import snapshot

import ouros
from ouros import AbstractOS, StatResult


class TestOS(AbstractOS):
    """A simple in-memory filesystem for testing."""

    __test__ = False

    def __init__(self) -> None:
        self.files: dict[str, bytes] = {}
        self.directories: set[str] = {'/'}

    def _ensure_parent_exists(self, path: str) -> None:
        """Ensure all parent directories exist."""
        parts = path.rstrip('/').split('/')
        for i in range(1, len(parts)):
            parent = '/'.join(parts[:i]) or '/'
            self.directories.add(parent)

    def path_exists(self, path: PurePosixPath) -> bool:
        p = str(path)
        return p in self.files or p in self.directories

    def path_is_file(self, path: PurePosixPath) -> bool:
        return str(path) in self.files

    def path_is_dir(self, path: PurePosixPath) -> bool:
        return str(path) in self.directories

    def path_is_symlink(self, path: PurePosixPath) -> bool:
        return False  # No symlink support in this simple implementation

    def path_read_text(self, path: PurePosixPath) -> str:
        p = str(path)
        if p not in self.files:
            raise FileNotFoundError(f'No such file: {p}')
        return self.files[p].decode('utf-8')

    def path_read_bytes(self, path: PurePosixPath) -> bytes:
        p = str(path)
        if p not in self.files:
            raise FileNotFoundError(f'No such file: {p}')
        return self.files[p]

    def path_write_text(self, path: PurePosixPath, data: str) -> int:
        p = str(path)
        self._ensure_parent_exists(p)
        self.files[p] = data.encode('utf-8')
        return len(data)

    def path_write_bytes(self, path: PurePosixPath, data: bytes) -> int:
        p = str(path)
        self._ensure_parent_exists(p)
        self.files[p] = data
        return len(data)

    def path_mkdir(self, path: PurePosixPath, parents: bool, exist_ok: bool) -> None:
        p = str(path)
        if p in self.directories:
            if not exist_ok:
                raise FileExistsError(f'Directory exists: {p}')
            return
        if parents:
            self._ensure_parent_exists(p)
        self.directories.add(p)

    def path_unlink(self, path: PurePosixPath) -> None:
        p = str(path)
        if p not in self.files:
            raise FileNotFoundError(f'No such file: {p}')
        del self.files[p]

    def path_rmdir(self, path: PurePosixPath) -> None:
        p = str(path)
        if p not in self.directories:
            raise FileNotFoundError(f'No such directory: {p}')
        # Check if directory is empty
        for f in self.files:
            if f.startswith(p + '/'):
                raise OSError(f'Directory not empty: {p}')
        for d in self.directories:
            if d != p and d.startswith(p + '/'):
                raise OSError(f'Directory not empty: {p}')
        self.directories.remove(p)

    def path_iterdir(self, path: PurePosixPath) -> list[PurePosixPath]:
        p = str(path)
        if p not in self.directories:
            raise FileNotFoundError(f'No such directory: {p}')
        result: list[PurePosixPath] = []
        prefix = p.rstrip('/') + '/'
        seen: set[str] = set()
        for f in self.files:
            if f.startswith(prefix):
                # Get immediate child name
                rest = f[len(prefix) :]
                child = rest.split('/')[0]
                if child and child not in seen:
                    seen.add(child)
                    result.append(PurePosixPath(prefix + child))
        for d in self.directories:
            if d.startswith(prefix) and d != p:
                rest = d[len(prefix) :]
                child = rest.split('/')[0]
                if child and child not in seen:
                    seen.add(child)
                    result.append(PurePosixPath(prefix + child))
        return sorted(result)

    def path_stat(self, path: PurePosixPath) -> StatResult:
        p = str(path)
        if p in self.files:
            return StatResult.file_stat(len(self.files[p]), 0o644, 0.0)
        elif p in self.directories:
            return StatResult.dir_stat(0o755, 0.0)
        else:
            raise FileNotFoundError(f'No such file or directory: {p}')

    def path_rename(self, path: PurePosixPath, target: PurePosixPath) -> None:
        p = str(path)
        t = str(target)
        if p in self.files:
            self._ensure_parent_exists(t)
            self.files[t] = self.files.pop(p)
        elif p in self.directories:
            self._ensure_parent_exists(t)
            self.directories.remove(p)
            self.directories.add(t)
            # Move all files under this directory
            prefix = p.rstrip('/') + '/'
            to_move = [(f, t + f[len(p) :]) for f in self.files if f.startswith(prefix)]
            for old, new in to_move:
                self.files[new] = self.files.pop(old)
        else:
            raise FileNotFoundError(f'No such file or directory: {p}')

    def path_resolve(self, path: PurePosixPath) -> str:
        # Simple implementation: just normalize the path
        p = str(path)
        parts: list[str] = []
        for part in p.split('/'):
            if part == '..':
                if parts:
                    parts.pop()
            elif part and part != '.':
                parts.append(part)
        return '/' + '/'.join(parts)

    def path_absolute(self, path: PurePosixPath) -> str:
        p = str(path)
        if p.startswith('/'):
            return p
        return '/' + p

    def getenv(self, key: str, default: str | None = None) -> str | None:
        # Simple virtual environment for testing
        env = {
            'TEST_VAR': 'test_value',
            'HOME': '/test/home',
        }
        return env.get(key, default)

    def get_environ(self) -> dict[str, str]:
        return {
            'TEST_VAR': 'test_value',
            'HOME': '/test/home',
        }


# =============================================================================
# Basic AbstractFileSystem tests
# =============================================================================


def test_abstract_filesystem_exists():
    """AbstractFileSystem.path_exists() works with os."""
    fs = TestOS()
    fs.files['/test.txt'] = b'hello'

    m = ouros.Sandbox('from pathlib import Path; Path("/test.txt").exists()')
    result = m.run(os=fs)

    assert result is True


def test_abstract_filesystem_exists_missing():
    """AbstractFileSystem.path_exists() returns False for missing files."""
    fs = TestOS()

    m = ouros.Sandbox('from pathlib import Path; Path("/missing.txt").exists()')
    result = m.run(os=fs)

    assert result is False


def test_abstract_filesystem_is_file():
    """AbstractFileSystem.path_is_file() distinguishes files from directories."""
    fs = TestOS()
    fs.files['/file.txt'] = b'content'
    fs.directories.add('/mydir')

    code = """
from pathlib import Path
(Path('/file.txt').is_file(), Path('/mydir').is_file())
"""
    m = ouros.Sandbox(code)
    result = m.run(os=fs)

    assert result == snapshot((True, False))


def test_abstract_filesystem_is_dir():
    """AbstractFileSystem.path_is_dir() distinguishes directories from files."""
    fs = TestOS()
    fs.files['/file.txt'] = b'content'
    fs.directories.add('/mydir')

    code = """
from pathlib import Path
(Path('/file.txt').is_dir(), Path('/mydir').is_dir())
"""
    m = ouros.Sandbox(code)
    result = m.run(os=fs)

    assert result == snapshot((False, True))


def test_abstract_filesystem_read_text():
    """AbstractFileSystem.path_read_text() returns file contents."""
    fs = TestOS()
    fs.files['/hello.txt'] = b'Hello, World!'

    m = ouros.Sandbox('from pathlib import Path; Path("/hello.txt").read_text()')
    result = m.run(os=fs)

    assert result == snapshot('Hello, World!')


def test_abstract_filesystem_read_text_missing():
    """AbstractFileSystem.path_read_text() raises FileNotFoundError for missing files."""
    fs = TestOS()

    m = ouros.Sandbox('from pathlib import Path; Path("/missing.txt").read_text()')
    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        m.run(os=fs)
    assert str(exc_info.value) == snapshot('FileNotFoundError: No such file: /missing.txt')
    assert isinstance(exc_info.value.exception(), FileNotFoundError)


def test_abstract_filesystem_read_bytes():
    """AbstractFileSystem.path_read_bytes() returns raw bytes."""
    fs = TestOS()
    fs.files['/data.bin'] = b'\x00\x01\x02\x03'

    m = ouros.Sandbox('from pathlib import Path; Path("/data.bin").read_bytes()')
    result = m.run(os=fs)

    assert result == snapshot(b'\x00\x01\x02\x03')


# =============================================================================
# stat() tests
# =============================================================================


def test_abstract_filesystem_stat_file():
    """AbstractFileSystem.path_stat() returns stat result for files."""
    fs = TestOS()
    fs.files['/file.txt'] = b'hello world'

    code = """
from pathlib import Path
s = Path('/file.txt').stat()
(s.st_size, s.st_mode)
"""
    m = ouros.Sandbox(code)
    result = m.run(os=fs)

    assert result == snapshot((11, 0o100644))


def test_abstract_filesystem_stat_directory():
    """AbstractFileSystem.path_stat() returns stat result for directories."""
    fs = TestOS()
    fs.directories.add('/mydir')

    code = """
from pathlib import Path
s = Path('/mydir').stat()
s.st_mode
"""
    m = ouros.Sandbox(code)
    result = m.run(os=fs)

    assert result == snapshot(0o040755)


def test_abstract_filesystem_stat_missing():
    """AbstractFileSystem.path_stat() raises FileNotFoundError for missing paths."""
    fs = TestOS()

    m = ouros.Sandbox('from pathlib import Path\nPath("/missing").stat()')
    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        m.run(os=fs)

    assert str(exc_info.value) == snapshot('FileNotFoundError: No such file or directory: /missing')
    assert exc_info.value.display() == snapshot("""\
Traceback (most recent call last):
  File "main.py", line 2, in <module>
    Path("/missing").stat()
    ~~~~~~~~~~~~~~~~~~~~~~~
FileNotFoundError: No such file or directory: /missing\
""")


# =============================================================================
# iterdir() tests
# =============================================================================


def test_abstract_filesystem_iterdir():
    """AbstractFileSystem.path_iterdir() lists directory contents."""
    fs = TestOS()
    fs.directories.add('/mydir')
    fs.files['/mydir/a.txt'] = b'a'
    fs.files['/mydir/b.txt'] = b'b'
    fs.directories.add('/mydir/subdir')

    code = """
from pathlib import Path
list(Path('/mydir').iterdir())
"""
    m = ouros.Sandbox(code)
    result = m.run(os=fs)

    # Result is a list of Path objects with child names joined to parent
    assert len(result) == 3
    names = sorted(str(p) for p in result)
    assert names == snapshot(['/mydir/a.txt', '/mydir/b.txt', '/mydir/subdir'])


def test_abstract_filesystem_iterdir_empty():
    """AbstractFileSystem.path_iterdir() returns empty list for empty directory."""
    fs = TestOS()
    fs.directories.add('/empty')

    code = """
from pathlib import Path
list(Path('/empty').iterdir())
"""
    m = ouros.Sandbox(code)
    result = m.run(os=fs)

    assert result == snapshot([])


# =============================================================================
# resolve() and absolute() tests
# =============================================================================


def test_abstract_filesystem_resolve():
    """AbstractFileSystem.path_resolve() normalizes paths."""
    fs = TestOS()

    code = """
from pathlib import Path
str(Path('/foo/bar/../baz').resolve())
"""
    m = ouros.Sandbox(code)
    result = m.run(os=fs)

    assert result == snapshot('/foo/baz')


def test_abstract_filesystem_absolute():
    """AbstractFileSystem.path_absolute() returns absolute path."""
    fs = TestOS()

    code = """
from pathlib import Path
str(Path('/already/absolute').absolute())
"""
    m = ouros.Sandbox(code)
    result = m.run(os=fs)

    assert result == snapshot('/already/absolute')


def test_abstract_filesystem_getenv():
    """AbstractFileSystem.getenv() returns environment variable value."""
    fs = TestOS()

    code = """
import os
os.getenv('TEST_VAR')
"""
    m = ouros.Sandbox(code)
    result = m.run(os=fs)

    assert result == snapshot('test_value')


def test_abstract_filesystem_getenv_missing():
    """AbstractFileSystem.getenv() returns None for missing variable."""
    fs = TestOS()

    code = """
import os
os.getenv('NONEXISTENT')
"""
    m = ouros.Sandbox(code)
    result = m.run(os=fs)

    assert result is None


def test_abstract_filesystem_getenv_default():
    """AbstractFileSystem.getenv() returns default for missing variable."""
    fs = TestOS()

    code = """
import os
os.getenv('NONEXISTENT', 'my_default')
"""
    m = ouros.Sandbox(code)
    result = m.run(os=fs)

    assert result == snapshot('my_default')


# =============================================================================
# file_stat / dir_stat helper tests
# =============================================================================


def test_file_stat_helper():
    """file_stat() creates a proper stat result."""
    stat = StatResult.file_stat(1024, 0o644, 1700000000.0)

    # Check it has the expected structure (10 fields)
    assert len(stat) == snapshot(10)
    # Index access: st_mode=0, st_size=6, st_mtime=8
    assert stat[0] == snapshot(0o100644)  # st_mode - file_stat adds file type bits
    assert stat[6] == snapshot(1024)  # st_size
    assert stat[8] == snapshot(1700000000.0)  # st_mtime


def test_dir_stat_helper():
    """dir_stat() creates a proper stat result for directories."""
    stat = StatResult.dir_stat(0o755, 1700000000.0)

    assert len(stat) == snapshot(10)
    # Index access: st_mode=0, st_size=6, st_mtime=8
    assert stat[0] == snapshot(0o040755)  # st_mode - dir_stat adds directory type bits
    assert stat[6] == snapshot(4096)  # st_size - directories have fixed size
    assert stat[8] == snapshot(1700000000.0)  # st_mtime


def test_path_ouros_to_py():
    m = ouros.Sandbox('from pathlib import Path; Path("/foo/bar/thing.txt")')
    result = m.run()
    assert result == PurePosixPath('/foo/bar/thing.txt')
    assert type(result) is PurePosixPath


def test_path_py_to_ouros():
    p = PurePosixPath('/foo/bar/thing.txt')
    m = ouros.Sandbox('f"type={type(p)} {p=}"', inputs=['p'])
    result = m.run(inputs={'p': p})
    assert result == snapshot("type=<class 'pathlib.PosixPath'> p=PosixPath('/foo/bar/thing.txt')")



# === iter() on various iterables ===
# iter() creates an iterator from an iterable

# iter() on list
it = iter([1, 2, 3])
assert next(it) == 1, 'iter list: first element should be 1'
assert next(it) == 2, 'iter list: second element should be 2'
assert next(it) == 3, 'iter list: third element should be 3'

# iter() on tuple
it = iter((10, 20))
assert next(it) == 10, 'iter tuple: first element should be 10'
assert next(it) == 20, 'iter tuple: second element should be 20'

# iter() on string
it = iter('ab')
assert next(it) == 'a', 'iter string: first element should be a'
assert next(it) == 'b', 'iter string: second element should be b'

# iter() on range
it = iter(range(3))
assert next(it) == 0, 'iter range: first element should be 0'
assert next(it) == 1, 'iter range: second element should be 1'
assert next(it) == 2, 'iter range: third element should be 2'

# iter() on dict iterates over keys
d = {'x': 1, 'y': 2}
it = iter(d)
keys = [next(it), next(it)]
assert 'x' in keys, 'iter dict: x should be in keys'
assert 'y' in keys, 'iter dict: y should be in keys'

# === next() with default value ===
# next() returns default when iterator is exhausted

it = iter([42])
assert next(it) == 42, 'next: first element should be 42'
assert next(it, 'done') == 'done', 'next with default: should return default when exhausted'

# Check default with various types
it = iter([])
assert next(it, None) is None, 'next with None default: should return None'
assert next(it, 0) == 0, 'next with 0 default: should return 0'
assert next(it, []) == [], 'next with empty list default: should return empty list'

# === iter() on iterator returns itself ===
# Calling iter() on an iterator should return the same iterator

original = iter([1, 2, 3])
same = iter(original)
# They should iterate over the same values
assert next(original) == 1, 'iter on iterator: original first should be 1'
assert next(same) == 2, 'iter on iterator: same should continue from 2'
assert next(original) == 3, 'iter on iterator: original should continue to 3'

# === Multiple independent iterators ===
# Creating multiple iterators over the same iterable should be independent

data = [1, 2, 3]
it1 = iter(data)
it2 = iter(data)
assert next(it1) == 1, 'independent iterators: it1 first should be 1'
assert next(it1) == 2, 'independent iterators: it1 second should be 2'
assert next(it2) == 1, 'independent iterators: it2 first should be 1 (independent)'
assert next(it1) == 3, 'independent iterators: it1 third should be 3'
assert next(it2) == 2, 'independent iterators: it2 second should be 2'



"""Tests for OSAccess class functionality.

These tests verify the OSAccess class behavior - the high-level virtual filesystem
that can be passed to Sandbox.run(os=...). Most tests run Python code through Sandbox
to verify behavior as it would be used in practice.

For tests of the AbstractOS interface via custom subclasses, see test_os_access_raw.py.
"""

from pathlib import PurePosixPath
from typing import Any

import pytest
from inline_snapshot import snapshot

from ouros import CallbackFile, MemoryFile, OSAccess, Sandbox, SandboxRuntimeError

# Alias for brevity in tests
P = PurePosixPath

# =============================================================================
# OSAccess Initialization & Validation
# =============================================================================


def test_non_absolute_path():
    """OSAccess rejects files with relative paths."""
    osa = OSAccess([MemoryFile('relative/path.txt', content='test')])
    assert osa.files[0].path.as_posix() == '/relative/path.txt'

    osa = OSAccess([MemoryFile('relative/path.txt', content='test')], root_dir='/foo/bar')
    assert osa.files[0].path.as_posix() == '/foo/bar/relative/path.txt'


def test_file_nested_within_file_rejected():
    """OSAccess rejects files nested within another file's path."""
    with pytest.raises(ValueError) as exc_info:
        OSAccess(
            [
                MemoryFile('/test/file.txt', content='outer'),
                MemoryFile('/test/file.txt/nested.txt', content='inner'),
            ]
        )
    assert str(exc_info.value) == snapshot(
        "Cannot put file MemoryFile(path=/test/file.txt/nested.txt, content='...', permissions=420) "
        "within sub-directory of file MemoryFile(path=/test/file.txt, content='...', permissions=420)"
    )


def test_empty_initialization():
    """OSAccess can be initialized with no files."""
    fs = OSAccess()
    result = Sandbox('from pathlib import Path; Path("/any/path").exists()').run(os=fs)
    assert result is False


def test_environ_parameter():
    """OSAccess accepts environ parameter for environment variables."""
    fs = OSAccess(environ={'MY_VAR': 'my_value'})
    result = Sandbox("import os; os.getenv('MY_VAR')").run(os=fs)
    assert result == snapshot('my_value')


# =============================================================================
# Path Existence Checks (via Sandbox)
# =============================================================================


def test_path_exists_file():
    """path_exists returns True for existing files."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello')])
    result = Sandbox('from pathlib import Path; Path("/test/file.txt").exists()').run(os=fs)
    assert result is True


def test_path_exists_directory():
    """path_exists returns True for directories created by file paths."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])
    result = Sandbox('from pathlib import Path; Path("/test/subdir").exists()').run(os=fs)
    assert result is True


def test_path_exists_nested():
    """path_exists handles deeply nested paths."""
    fs = OSAccess([MemoryFile('/a/b/c/d/file.txt', content='deep')])
    code = """
from pathlib import Path
(Path('/a').exists(), Path('/a/b').exists(), Path('/a/b/c').exists(), Path('/a/b/c/d').exists())
"""
    result = Sandbox(code).run(os=fs)
    assert result == snapshot((True, True, True, True))


def test_path_exists_missing():
    """path_exists returns False for non-existent paths."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello')])
    result = Sandbox('from pathlib import Path; Path("/other/path").exists()').run(os=fs)
    assert result is False


def test_path_is_file_for_file():
    """path_is_file returns True for files."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello')])
    result = Sandbox('from pathlib import Path; Path("/test/file.txt").is_file()').run(os=fs)
    assert result is True


def test_path_is_file_for_directory():
    """path_is_file returns False for directories."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])
    result = Sandbox('from pathlib import Path; Path("/test/subdir").is_file()').run(os=fs)
    assert result is False


def test_path_is_file_missing():
    """path_is_file returns False for non-existent paths."""
    fs = OSAccess()
    result = Sandbox('from pathlib import Path; Path("/missing").is_file()').run(os=fs)
    assert result is False


def test_path_is_dir_for_directory():
    """path_is_dir returns True for directories."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])
    result = Sandbox('from pathlib import Path; Path("/test/subdir").is_dir()').run(os=fs)
    assert result is True


def test_path_is_dir_for_file():
    """path_is_dir returns False for files."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello')])
    result = Sandbox('from pathlib import Path; Path("/test/file.txt").is_dir()').run(os=fs)
    assert result is False


def test_path_is_dir_missing():
    """path_is_dir returns False for non-existent paths."""
    fs = OSAccess()
    result = Sandbox('from pathlib import Path; Path("/missing").is_dir()').run(os=fs)
    assert result is False


def test_path_is_symlink_always_false():
    """path_is_symlink always returns False (no symlink support)."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello')])
    code = """
from pathlib import Path
(Path('/test/file.txt').is_symlink(), Path('/test').is_symlink(), Path('/missing').is_symlink())
"""
    result = Sandbox(code).run(os=fs)
    assert result == snapshot((False, False, False))


# =============================================================================
# Reading Files (via Sandbox)
# =============================================================================


def test_read_text_string_content():
    """path_read_text returns string content directly."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello world')])
    result = Sandbox('from pathlib import Path; Path("/test/file.txt").read_text()').run(os=fs)
    assert result == snapshot('hello world')


def test_read_text_bytes_content_decoded():
    """path_read_text decodes bytes content as UTF-8."""
    fs = OSAccess([MemoryFile('/test/file.txt', content=b'bytes content')])
    result = Sandbox('from pathlib import Path; Path("/test/file.txt").read_text()').run(os=fs)
    assert result == snapshot('bytes content')


def test_read_bytes_bytes_content():
    """path_read_bytes returns bytes content directly."""
    fs = OSAccess([MemoryFile('/test/file.bin', content=b'\x00\x01\x02\x03')])
    result = Sandbox('from pathlib import Path; Path("/test/file.bin").read_bytes()').run(os=fs)
    assert result == snapshot(b'\x00\x01\x02\x03')


def test_read_bytes_string_content_encoded():
    """path_read_bytes encodes string content as UTF-8."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello')])
    result = Sandbox('from pathlib import Path; Path("/test/file.txt").read_bytes()').run(os=fs)
    assert result == snapshot(b'hello')


def test_read_text_file_not_found():
    """path_read_text raises FileNotFoundError for missing files."""
    fs = OSAccess()
    with pytest.raises(SandboxRuntimeError) as exc_info:
        Sandbox('from pathlib import Path; Path("/missing.txt").read_text()').run(os=fs)
    assert str(exc_info.value) == snapshot("FileNotFoundError: [Errno 2] No such file or directory: '/missing.txt'")


def test_read_bytes_file_not_found():
    """path_read_bytes raises FileNotFoundError for missing files."""
    fs = OSAccess()
    with pytest.raises(SandboxRuntimeError) as exc_info:
        Sandbox('from pathlib import Path; Path("/missing.bin").read_bytes()').run(os=fs)
    assert str(exc_info.value) == snapshot("FileNotFoundError: [Errno 2] No such file or directory: '/missing.bin'")


def test_read_text_is_a_directory():
    """path_read_text raises error for directories."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])
    with pytest.raises(SandboxRuntimeError) as exc_info:
        Sandbox('from pathlib import Path; Path("/test/subdir").read_text()').run(os=fs)
    # Sandbox reports this as OSError, not IsADirectoryError
    assert str(exc_info.value) == snapshot("IsADirectoryError: [Errno 21] Is a directory: '/test/subdir'")


def test_read_bytes_is_a_directory():
    """path_read_bytes raises error for directories."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])
    with pytest.raises(SandboxRuntimeError) as exc_info:
        Sandbox('from pathlib import Path; Path("/test/subdir").read_bytes()').run(os=fs)
    # Sandbox reports this as OSError, not IsADirectoryError
    assert str(exc_info.value) == snapshot("IsADirectoryError: [Errno 21] Is a directory: '/test/subdir'")


# =============================================================================
# Writing Files (via Sandbox)
# =============================================================================


def test_write_text_via_ouros():
    """Path.write_text() creates a new file via Sandbox."""
    fs = OSAccess([MemoryFile('/test/existing.txt', content='existing')])

    code = """
from pathlib import Path
Path('/test/new.txt').write_text('new content')
"""
    result = Sandbox(code).run(os=fs)
    # write_text returns the number of bytes written
    assert result == snapshot(11)

    # Verify file was created
    assert fs.path_exists(P('/test/new.txt')) is True
    assert fs.path_read_text(P('/test/new.txt')) == 'new content'


def test_write_text_overwrite_via_ouros():
    """Path.write_text() overwrites existing file via Sandbox."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='original')])

    code = """
from pathlib import Path
Path('/test/file.txt').write_text('updated')
"""
    Sandbox(code).run(os=fs)
    assert fs.path_read_text(P('/test/file.txt')) == 'updated'


def test_write_bytes_via_ouros():
    """Path.write_bytes() creates a new file via Sandbox."""
    fs = OSAccess([MemoryFile('/test/existing.txt', content='existing')])

    code = """
from pathlib import Path
Path('/test/new.bin').write_bytes(b'binary data')
"""
    result = Sandbox(code).run(os=fs)
    assert result == snapshot(11)
    assert fs.path_read_bytes(P('/test/new.bin')) == b'binary data'


def test_write_text_parent_not_exists_via_ouros():
    """Path.write_text() raises FileNotFoundError when parent doesn't exist via Sandbox."""
    fs = OSAccess()
    with pytest.raises(SandboxRuntimeError) as exc_info:
        Sandbox("from pathlib import Path; Path('/no/parent/file.txt').write_text('test')").run(os=fs)
    assert str(exc_info.value) == snapshot(
        "FileNotFoundError: [Errno 2] No such file or directory: '/no/parent/file.txt'"
    )


def test_write_text_to_directory_via_ouros():
    """Path.write_text() raises IsADirectoryError when writing to a directory via Sandbox."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])
    with pytest.raises(SandboxRuntimeError) as exc_info:
        Sandbox("from pathlib import Path; Path('/test/subdir').write_text('test')").run(os=fs)
    assert str(exc_info.value) == snapshot("IsADirectoryError: [Errno 21] Is a directory: '/test/subdir'")


# =============================================================================
# Writing Files (via direct API)
# =============================================================================


def test_write_text_new_file_direct():
    """path_write_text creates a new file via direct API."""
    fs = OSAccess([MemoryFile('/test/existing.txt', content='existing')])

    # Write a new file
    fs.path_write_text(P('/test/new.txt'), 'new content')

    # Verify it was created
    assert fs.path_exists(P('/test/new.txt')) is True
    assert fs.path_read_text(P('/test/new.txt')) == 'new content'


def test_write_text_overwrite_existing_direct():
    """path_write_text overwrites existing file content via direct API."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='original')])

    fs.path_write_text(P('/test/file.txt'), 'updated')
    assert fs.path_read_text(P('/test/file.txt')) == 'updated'


def test_write_bytes_new_file_direct():
    """path_write_bytes creates a new file via direct API."""
    fs = OSAccess([MemoryFile('/test/existing.txt', content='existing')])

    fs.path_write_bytes(P('/test/new.bin'), b'binary data')
    assert fs.path_read_bytes(P('/test/new.bin')) == b'binary data'


def test_write_bytes_overwrite_existing_direct():
    """path_write_bytes overwrites existing file content via direct API."""
    fs = OSAccess([MemoryFile('/test/file.bin', content=b'original')])

    fs.path_write_bytes(P('/test/file.bin'), b'updated')
    assert fs.path_read_bytes(P('/test/file.bin')) == b'updated'


def test_write_text_parent_not_exists_direct():
    """path_write_text raises FileNotFoundError when parent doesn't exist via direct API."""
    fs = OSAccess()
    with pytest.raises(FileNotFoundError) as exc_info:
        fs.path_write_text(P('/no/parent/file.txt'), 'test')
    assert str(exc_info.value) == snapshot("[Errno 2] No such file or directory: '/no/parent/file.txt'")


def test_write_text_to_directory_direct():
    """path_write_text raises IsADirectoryError when writing to a directory via direct API."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])
    with pytest.raises(IsADirectoryError) as exc_info:
        fs.path_write_text(P('/test/subdir'), 'test')
    assert str(exc_info.value) == snapshot("[Errno 21] Is a directory: '/test/subdir'")


# =============================================================================
# Directory Operations - mkdir (via Sandbox)
# =============================================================================


def test_mkdir_basic_via_ouros():
    """Path.mkdir() creates a directory via Sandbox."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello')])

    code = """
from pathlib import Path
Path('/test/newdir').mkdir()
"""
    Sandbox(code).run(os=fs)
    assert fs.path_is_dir(P('/test/newdir')) is True


def test_mkdir_with_parents_via_ouros():
    """Path.mkdir(parents=True) creates parent directories via Sandbox."""
    fs = OSAccess()

    code = """
from pathlib import Path
Path('/a/b/c/d').mkdir(parents=True)
"""
    Sandbox(code).run(os=fs)
    assert fs.path_is_dir(P('/a')) is True
    assert fs.path_is_dir(P('/a/b')) is True
    assert fs.path_is_dir(P('/a/b/c')) is True
    assert fs.path_is_dir(P('/a/b/c/d')) is True


def test_mkdir_exist_ok_true_via_ouros():
    """Path.mkdir(exist_ok=True) doesn't raise for existing directory via Sandbox."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])

    code = """
from pathlib import Path
Path('/test/subdir').mkdir(exist_ok=True)
"""
    # Should not raise
    Sandbox(code).run(os=fs)
    assert fs.path_is_dir(P('/test/subdir')) is True


def test_mkdir_exist_ok_false_via_ouros():
    """Path.mkdir() raises OSError (FileExistsError) for existing directory via Sandbox."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])

    with pytest.raises(SandboxRuntimeError) as exc_info:
        Sandbox("from pathlib import Path; Path('/test/subdir').mkdir()").run(os=fs)
    # Sandbox maps FileExistsError to OSError
    assert str(exc_info.value) == snapshot("FileExistsError: [Errno 17] File exists: '/test/subdir'")


def test_mkdir_parent_not_exists_via_ouros():
    """Path.mkdir() raises FileNotFoundError when parent doesn't exist via Sandbox."""
    fs = OSAccess()

    with pytest.raises(SandboxRuntimeError) as exc_info:
        Sandbox("from pathlib import Path; Path('/no/parent/dir').mkdir()").run(os=fs)
    assert str(exc_info.value) == snapshot("FileNotFoundError: [Errno 2] No such file or directory: '/no/parent/dir'")


# =============================================================================
# Directory Operations - mkdir (via direct API)
# =============================================================================


def test_mkdir_basic_direct():
    """path_mkdir creates a directory via direct API."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello')])

    fs.path_mkdir(P('/test/newdir'), parents=False, exist_ok=False)
    assert fs.path_is_dir(P('/test/newdir')) is True


def test_mkdir_with_parents_direct():
    """path_mkdir with parents=True creates parent directories via direct API."""
    fs = OSAccess()

    fs.path_mkdir(P('/a/b/c/d'), parents=True, exist_ok=False)
    assert fs.path_is_dir(P('/a')) is True
    assert fs.path_is_dir(P('/a/b')) is True
    assert fs.path_is_dir(P('/a/b/c')) is True
    assert fs.path_is_dir(P('/a/b/c/d')) is True


def test_mkdir_exist_ok_true_direct():
    """path_mkdir with exist_ok=True doesn't raise for existing directory via direct API."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])

    # Should not raise
    fs.path_mkdir(P('/test/subdir'), parents=False, exist_ok=True)
    assert fs.path_is_dir(P('/test/subdir')) is True


def test_mkdir_exist_ok_false_direct():
    """path_mkdir with exist_ok=False raises for existing directory via direct API."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])

    with pytest.raises(FileExistsError) as exc_info:
        fs.path_mkdir(P('/test/subdir'), parents=False, exist_ok=False)
    assert str(exc_info.value) == snapshot("[Errno 17] File exists: '/test/subdir'")


def test_mkdir_file_exists_direct():
    """path_mkdir raises FileExistsError when a file exists at the path via direct API."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello')])

    with pytest.raises(FileExistsError) as exc_info:
        fs.path_mkdir(P('/test/file.txt'), parents=False, exist_ok=False)
    assert str(exc_info.value) == snapshot("[Errno 17] File exists: '/test/file.txt'")


def test_mkdir_parent_not_exists_direct():
    """path_mkdir without parents raises FileNotFoundError when parent doesn't exist via direct API."""
    fs = OSAccess()

    with pytest.raises(FileNotFoundError) as exc_info:
        fs.path_mkdir(P('/no/parent/dir'), parents=False, exist_ok=False)
    assert str(exc_info.value) == snapshot("[Errno 2] No such file or directory: '/no/parent/dir'")


def test_mkdir_parent_is_file_direct():
    """path_mkdir raises NotADirectoryError when parent is a file via direct API."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello')])

    with pytest.raises(NotADirectoryError) as exc_info:
        fs.path_mkdir(P('/test/file.txt/subdir'), parents=True, exist_ok=False)
    assert str(exc_info.value) == snapshot("[Errno 20] Not a directory: '/test/file.txt/subdir'")


# =============================================================================
# Directory Operations - rmdir (via Sandbox)
# =============================================================================


def test_rmdir_empty_directory_via_ouros():
    """Path.rmdir() removes an empty directory via Sandbox."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])
    fs.path_mkdir(P('/test/newdir'), parents=False, exist_ok=False)

    code = """
from pathlib import Path
Path('/test/newdir').rmdir()
"""
    Sandbox(code).run(os=fs)
    assert fs.path_exists(P('/test/newdir')) is False


def test_rmdir_non_empty_directory_via_ouros():
    """Path.rmdir() raises OSError for non-empty directory via Sandbox."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])

    with pytest.raises(SandboxRuntimeError) as exc_info:
        Sandbox("from pathlib import Path; Path('/test/subdir').rmdir()").run(os=fs)
    assert str(exc_info.value) == snapshot("OSError: [Errno 39] Directory not empty: '/test/subdir'")


def test_rmdir_not_found_via_ouros():
    """Path.rmdir() raises FileNotFoundError for non-existent path via Sandbox."""
    fs = OSAccess()

    with pytest.raises(SandboxRuntimeError) as exc_info:
        Sandbox("from pathlib import Path; Path('/missing').rmdir()").run(os=fs)
    assert str(exc_info.value) == snapshot("FileNotFoundError: [Errno 2] No such file or directory: '/missing'")


def test_rmdir_file_not_directory_via_ouros():
    """Path.rmdir() raises NotADirectoryError for files via Sandbox."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello')])

    with pytest.raises(SandboxRuntimeError) as exc_info:
        Sandbox("from pathlib import Path; Path('/test/file.txt').rmdir()").run(os=fs)
    assert str(exc_info.value) == snapshot("NotADirectoryError: [Errno 20] Not a directory: '/test/file.txt'")


# =============================================================================
# Directory Operations - rmdir (via direct API)
# =============================================================================


def test_rmdir_empty_directory_direct():
    """path_rmdir removes an empty directory via direct API."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])

    fs.path_mkdir(P('/test/newdir'), parents=False, exist_ok=False)
    fs.path_rmdir(P('/test/newdir'))
    assert fs.path_exists(P('/test/newdir')) is False


def test_rmdir_non_empty_directory_direct():
    """path_rmdir raises OSError for non-empty directory via direct API."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])

    with pytest.raises(OSError) as exc_info:
        fs.path_rmdir(P('/test/subdir'))
    assert str(exc_info.value) == snapshot("[Errno 39] Directory not empty: '/test/subdir'")


def test_rmdir_file_not_directory_direct():
    """path_rmdir raises NotADirectoryError for files via direct API."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello')])

    with pytest.raises(NotADirectoryError) as exc_info:
        fs.path_rmdir(P('/test/file.txt'))
    assert str(exc_info.value) == snapshot("[Errno 20] Not a directory: '/test/file.txt'")


def test_rmdir_not_found_direct():
    """path_rmdir raises FileNotFoundError for non-existent path via direct API."""
    fs = OSAccess()

    with pytest.raises(FileNotFoundError) as exc_info:
        fs.path_rmdir(P('/missing'))
    assert str(exc_info.value) == snapshot("[Errno 2] No such file or directory: '/missing'")


# =============================================================================
# Directory Operations - iterdir (via Sandbox)
# =============================================================================


def test_iterdir_list_contents():
    """path_iterdir lists directory contents."""
    fs = OSAccess(
        [
            MemoryFile('/test/a.txt', content='a'),
            MemoryFile('/test/b.txt', content='b'),
            MemoryFile('/test/subdir/c.txt', content='c'),
        ]
    )
    code = """
from pathlib import Path
[str(p) for p in Path('/test').iterdir()]
"""
    result = Sandbox(code).run(os=fs)
    # Result may be in any order, so sort in Python
    assert sorted(result) == snapshot(['/test/a.txt', '/test/b.txt', '/test/subdir'])


def test_iterdir_empty_directory_direct():
    """path_iterdir returns empty list for empty directory via direct API."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])
    fs.path_mkdir(P('/test/empty'), parents=False, exist_ok=False)

    result = fs.path_iterdir(P('/test/empty'))
    assert result == snapshot([])


def test_iterdir_not_a_directory_direct():
    """path_iterdir raises NotADirectoryError for files via direct API."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello')])

    with pytest.raises(NotADirectoryError) as exc_info:
        fs.path_iterdir(P('/test/file.txt'))
    assert str(exc_info.value) == snapshot("[Errno 20] Not a directory: '/test/file.txt'")


def test_iterdir_not_found():
    """path_iterdir raises FileNotFoundError for non-existent path."""
    fs = OSAccess()
    with pytest.raises(SandboxRuntimeError) as exc_info:
        Sandbox("from pathlib import Path; list(Path('/missing').iterdir())").run(os=fs)
    assert str(exc_info.value) == snapshot("FileNotFoundError: [Errno 2] No such file or directory: '/missing'")


# =============================================================================
# File Operations - unlink (via Sandbox)
# =============================================================================


def test_unlink_file_via_ouros():
    """Path.unlink() removes a file via Sandbox."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello')])

    code = """
from pathlib import Path
Path('/test/file.txt').unlink()
"""
    Sandbox(code).run(os=fs)
    assert fs.path_exists(P('/test/file.txt')) is False


def test_unlink_file_not_found_via_ouros():
    """Path.unlink() raises FileNotFoundError for non-existent files via Sandbox."""
    fs = OSAccess()

    with pytest.raises(SandboxRuntimeError) as exc_info:
        Sandbox("from pathlib import Path; Path('/missing.txt').unlink()").run(os=fs)
    assert str(exc_info.value) == snapshot("FileNotFoundError: [Errno 2] No such file or directory: '/missing.txt'")


def test_unlink_is_directory_via_ouros():
    """Path.unlink() raises IsADirectoryError for directories via Sandbox."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])

    with pytest.raises(SandboxRuntimeError) as exc_info:
        Sandbox("from pathlib import Path; Path('/test/subdir').unlink()").run(os=fs)
    assert str(exc_info.value) == snapshot("IsADirectoryError: [Errno 21] Is a directory: '/test/subdir'")


# =============================================================================
# File Operations - unlink (via direct API)
# =============================================================================


def test_unlink_file_direct():
    """path_unlink removes a file via direct API."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello')])

    fs.path_unlink(P('/test/file.txt'))
    assert fs.path_exists(P('/test/file.txt')) is False


def test_unlink_file_not_found_direct():
    """path_unlink raises FileNotFoundError for non-existent files via direct API."""
    fs = OSAccess()

    with pytest.raises(FileNotFoundError) as exc_info:
        fs.path_unlink(P('/missing.txt'))
    assert str(exc_info.value) == snapshot("[Errno 2] No such file or directory: '/missing.txt'")


def test_unlink_is_directory_direct():
    """path_unlink raises IsADirectoryError for directories via direct API."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])

    with pytest.raises(IsADirectoryError) as exc_info:
        fs.path_unlink(P('/test/subdir'))
    assert str(exc_info.value) == snapshot("[Errno 21] Is a directory: '/test/subdir'")


# =============================================================================
# Stat Operations (via Sandbox)
# =============================================================================


def test_stat_file():
    """path_stat returns stat result for files with size and mode."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello world')])
    code = """
from pathlib import Path
s = Path('/test/file.txt').stat()
(s.st_size, s.st_mode & 0o777)
"""
    result = Sandbox(code).run(os=fs)
    assert result == snapshot((11, 0o644))


def test_stat_file_custom_permissions():
    """path_stat returns custom file permissions."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello', permissions=0o755)])
    code = """
from pathlib import Path
s = Path('/test/file.txt').stat()
s.st_mode & 0o777
"""
    result = Sandbox(code).run(os=fs)
    assert result == snapshot(0o755)


def test_stat_directory():
    """path_stat returns stat result for directories."""
    fs = OSAccess([MemoryFile('/test/subdir/file.txt', content='hello')])
    code = """
from pathlib import Path
s = Path('/test/subdir').stat()
s.st_mode
"""
    result = Sandbox(code).run(os=fs)
    # Directory mode bits: 0o040000 (directory) | 0o755 (default perms) = 0o040755
    assert result == snapshot(0o040755)


def test_stat_file_not_found():
    """path_stat raises FileNotFoundError for non-existent paths."""
    fs = OSAccess()
    with pytest.raises(SandboxRuntimeError) as exc_info:
        Sandbox("from pathlib import Path; Path('/missing').stat()").run(os=fs)
    assert str(exc_info.value) == snapshot("FileNotFoundError: [Errno 2] No such file or directory: '/missing'")


def test_stat_bytes_content_size():
    """path_stat calculates size correctly for bytes content."""
    fs = OSAccess([MemoryFile('/test/file.bin', content=b'\x00\x01\x02\x03\x04')])
    code = """
from pathlib import Path
Path('/test/file.bin').stat().st_size
"""
    result = Sandbox(code).run(os=fs)
    assert result == snapshot(5)


def test_stat_unicode_size():
    """path_stat calculates size as encoded UTF-8 bytes for string content."""
    # Unicode snowman is 3 bytes in UTF-8
    fs = OSAccess([MemoryFile('/test/file.txt', content='☃')])
    code = """
from pathlib import Path
Path('/test/file.txt').stat().st_size
"""
    result = Sandbox(code).run(os=fs)
    assert result == snapshot(3)


# =============================================================================
# Rename Operations (via Sandbox)
# =============================================================================


def test_rename_file_via_ouros():
    """Path.rename() renames a file via Sandbox."""
    fs = OSAccess([MemoryFile('/test/old.txt', content='content')])

    code = """
from pathlib import Path
Path('/test/old.txt').rename(Path('/test/new.txt'))
"""
    Sandbox(code).run(os=fs)

    assert fs.path_exists(P('/test/old.txt')) is False
    assert fs.path_exists(P('/test/new.txt')) is True
    assert fs.path_read_text(P('/test/new.txt')) == 'content'


def test_rename_source_not_found_via_ouros():
    """Path.rename() raises FileNotFoundError when source doesn't exist via Sandbox."""
    fs = OSAccess()

    with pytest.raises(SandboxRuntimeError) as exc_info:
        Sandbox("from pathlib import Path; Path('/missing.txt').rename(Path('/new.txt'))").run(os=fs)
    assert str(exc_info.value) == snapshot(
        "FileNotFoundError: [Errno 2] No such file or directory: '/missing.txt' -> '/new.txt'"
    )


def test_rename_target_parent_not_found_via_ouros():
    """Path.rename() raises FileNotFoundError when target parent doesn't exist via Sandbox."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='content')])

    with pytest.raises(SandboxRuntimeError) as exc_info:
        Sandbox("from pathlib import Path; Path('/test/file.txt').rename(Path('/no/parent/file.txt'))").run(os=fs)
    assert str(exc_info.value) == snapshot(
        "FileNotFoundError: [Errno 2] No such file or directory: '/test/file.txt' -> '/no/parent/file.txt'"
    )


# =============================================================================
# Rename Operations (via direct API)
# =============================================================================


def test_rename_file_direct():
    """path_rename renames a file via direct API."""
    fs = OSAccess([MemoryFile('/test/old.txt', content='content')])

    fs.path_rename(P('/test/old.txt'), P('/test/new.txt'))

    assert fs.path_exists(P('/test/old.txt')) is False
    assert fs.path_exists(P('/test/new.txt')) is True
    assert fs.path_read_text(P('/test/new.txt')) == 'content'


def test_rename_source_not_found_direct():
    """path_rename raises FileNotFoundError when source doesn't exist via direct API."""
    fs = OSAccess()

    with pytest.raises(FileNotFoundError) as exc_info:
        fs.path_rename(P('/missing.txt'), P('/new.txt'))
    assert str(exc_info.value) == snapshot("[Errno 2] No such file or directory: '/missing.txt' -> '/new.txt'")


def test_rename_target_parent_not_found_direct():
    """path_rename raises FileNotFoundError when target parent doesn't exist via direct API."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='content')])

    with pytest.raises(FileNotFoundError) as exc_info:
        fs.path_rename(P('/test/file.txt'), P('/no/parent/file.txt'))
    assert str(exc_info.value) == snapshot(
        "[Errno 2] No such file or directory: '/test/file.txt' -> '/no/parent/file.txt'"
    )


def test_rename_directory_direct():
    """path_rename renames a directory via direct API."""
    fs = OSAccess([MemoryFile('/test/olddir/file.txt', content='content')])
    fs.path_mkdir(P('/test/newdir'), parents=False, exist_ok=False)

    fs.path_rename(P('/test/newdir'), P('/test/renamed'))
    assert fs.path_is_dir(P('/test/renamed')) is True


def test_rename_directory_non_empty_target_direct():
    """path_rename raises OSError when renaming directory to non-empty target via direct API."""
    fs = OSAccess(
        [
            MemoryFile('/test/src/a.txt', content='a'),
            MemoryFile('/test/dst/b.txt', content='b'),
        ]
    )

    with pytest.raises(OSError) as exc_info:
        fs.path_rename(P('/test/src'), P('/test/dst'))
    assert str(exc_info.value) == snapshot("[Errno 66] Directory not empty: '/test/src' -> '/test/dst'")


def test_rename_directory_updates_file_paths_direct():
    """path_rename updates paths of all files within renamed directory."""
    file1 = MemoryFile('/old/dir/file1.txt', content='one')
    file2 = MemoryFile('/old/dir/subdir/file2.txt', content='two')
    fs = OSAccess([file1, file2])

    # Create target parent and rename the directory
    fs.path_mkdir(P('/new'), parents=False, exist_ok=False)
    fs.path_rename(P('/old/dir'), P('/new/location'))

    # Verify files are accessible at new paths
    assert fs.path_read_text(P('/new/location/file1.txt')) == 'one'
    assert fs.path_read_text(P('/new/location/subdir/file2.txt')) == 'two'

    # Verify the AbstractFile objects have updated paths
    assert file1.path.as_posix() == '/new/location/file1.txt'
    assert file2.path.as_posix() == '/new/location/subdir/file2.txt'

    # Verify old paths no longer exist
    assert fs.path_exists(P('/old/dir')) is False
    assert fs.path_exists(P('/old/dir/file1.txt')) is False


# =============================================================================
# Path Resolution (via Sandbox)
# =============================================================================


def test_path_resolve_absolute():
    """path_resolve returns absolute path."""
    fs = OSAccess([MemoryFile('/test/file.txt', content='hello')])
    code = """
from pathlib import Path
str(Path('/test/file.txt').resolve())
"""
    result = Sandbox(code).run(os=fs)
    assert result == snapshot('/test/file.txt')


def test_path_absolute_already_absolute():
    """path_absolute returns same path for already absolute path."""
    fs = OSAccess()
    code = """
from pathlib import Path
str(Path('/already/absolute').absolute())
"""
    result = Sandbox(code).run(os=fs)
    assert result == snapshot('/already/absolute')


def test_path_absolute_relative():
    """path_absolute converts relative path to absolute."""
    fs = OSAccess()
    code = """
from pathlib import Path
str(Path('relative/path').absolute())
"""
    result = Sandbox(code).run(os=fs)
    assert result == snapshot('/relative/path')


def test_path_resolve_same_as_absolute():
    """path_resolve behaves same as absolute (no symlinks in OSAccess)."""
    fs = OSAccess()
    code = """
from pathlib import Path
str(Path('relative').resolve()) == str(Path('relative').absolute())
"""
    result = Sandbox(code).run(os=fs)
    assert result is True


# =============================================================================
# Environment Variables (via Sandbox)
# =============================================================================


def test_getenv_existing_key():
    """getenv returns value for existing key."""
    fs = OSAccess(environ={'MY_VAR': 'my_value'})
    result = Sandbox("import os; os.getenv('MY_VAR')").run(os=fs)
    assert result == snapshot('my_value')


def test_getenv_missing_key():
    """getenv returns None for missing key."""
    fs = OSAccess(environ={'OTHER': 'value'})
    result = Sandbox("import os; os.getenv('MISSING')").run(os=fs)
    assert result is None


def test_getenv_missing_with_default():
    """getenv returns default for missing key when default provided."""
    fs = OSAccess(environ={})
    result = Sandbox("import os; os.getenv('MISSING', 'default_value')").run(os=fs)
    assert result == snapshot('default_value')


def test_getenv_multiple_vars():
    """getenv handles multiple environment variables."""
    fs = OSAccess(environ={'VAR1': 'value1', 'VAR2': 'value2', 'VAR3': 'value3'})
    code = """
import os
(os.getenv('VAR1'), os.getenv('VAR2'), os.getenv('VAR3'))
"""
    result = Sandbox(code).run(os=fs)
    assert result == snapshot(('value1', 'value2', 'value3'))


def test_get_environ_returns_dict():
    """os.environ returns the full environ dict."""
    fs = OSAccess(environ={'HOME': '/home/user', 'USER': 'testuser'})
    result = Sandbox('import os; os.environ').run(os=fs)
    assert result == snapshot({'HOME': '/home/user', 'USER': 'testuser'})


def test_get_environ_key_access():
    """os.environ['KEY'] returns the value."""
    fs = OSAccess(environ={'MY_VAR': 'my_value'})
    result = Sandbox("import os; os.environ['MY_VAR']").run(os=fs)
    assert result == snapshot('my_value')


def test_get_environ_key_missing_raises():
    """os.environ['MISSING'] raises KeyError."""
    fs = OSAccess(environ={})
    with pytest.raises(SandboxRuntimeError) as exc_info:
        Sandbox("import os; os.environ['MISSING']").run(os=fs)
    assert str(exc_info.value) == snapshot('KeyError: MISSING')


def test_get_environ_get_method():
    """os.environ.get() works correctly."""
    fs = OSAccess(environ={'HOME': '/home/user'})
    result = Sandbox("import os; os.environ.get('HOME')").run(os=fs)
    assert result == snapshot('/home/user')


def test_get_environ_get_missing_with_default():
    """os.environ.get() returns default for missing key."""
    fs = OSAccess(environ={})
    result = Sandbox("import os; os.environ.get('MISSING', 'fallback')").run(os=fs)
    assert result == snapshot('fallback')


def test_get_environ_len():
    """len(os.environ) returns the number of env vars."""
    fs = OSAccess(environ={'A': '1', 'B': '2', 'C': '3'})
    result = Sandbox('import os; len(os.environ)').run(os=fs)
    assert result == snapshot(3)


def test_get_environ_contains():
    """'KEY' in os.environ tests membership."""
    fs = OSAccess(environ={'PRESENT': 'value'})
    code = """
import os
('PRESENT' in os.environ, 'ABSENT' in os.environ)
"""
    result = Sandbox(code).run(os=fs)
    assert result == snapshot((True, False))


def test_get_environ_keys():
    """os.environ.keys() returns the keys."""
    fs = OSAccess(environ={'X': '1', 'Y': '2'})
    result = Sandbox('import os; list(os.environ.keys())').run(os=fs)
    assert set(result) == snapshot({'X', 'Y'})


def test_get_environ_values():
    """os.environ.values() returns the values."""
    fs = OSAccess(environ={'X': 'a', 'Y': 'b'})
    result = Sandbox('import os; list(os.environ.values())').run(os=fs)
    assert set(result) == snapshot({'a', 'b'})


def test_get_environ_items():
    """os.environ.items() returns key-value pairs."""
    fs = OSAccess(environ={'X': '1', 'Y': '2'})
    result = Sandbox('import os; list(os.environ.items())').run(os=fs)
    assert set(result) == snapshot({('X', '1'), ('Y', '2')})


def test_get_environ_empty():
    """os.environ returns empty dict when no environ provided."""
    fs = OSAccess()
    result = Sandbox('import os; os.environ').run(os=fs)
    assert result == snapshot({})


# =============================================================================
# MemoryFile Behavior
# =============================================================================


def test_memory_file_string_content():
    """MemoryFile stores and returns string content."""
    file = MemoryFile('/test/file.txt', content='hello')
    assert file.read_content() == snapshot('hello')
    assert file.path.as_posix() == snapshot('/test/file.txt')
    assert file.name == snapshot('file.txt')


def test_memory_file_bytes_content():
    """MemoryFile stores and returns bytes content."""
    file = MemoryFile('/test/file.bin', content=b'\x00\x01\x02')
    assert file.read_content() == snapshot(b'\x00\x01\x02')


def test_memory_file_custom_permissions():
    """MemoryFile accepts custom permissions."""
    file = MemoryFile('/test/exec.sh', content='#!/bin/bash', permissions=0o755)
    assert file.permissions == snapshot(0o755)


def test_memory_file_write_and_read():
    """MemoryFile supports writing and re-reading content."""
    file = MemoryFile('/test/file.txt', content='original')
    file.write_content('updated')
    assert file.read_content() == snapshot('updated')


def test_memory_file_delete():
    """MemoryFile can be marked as deleted."""
    file = MemoryFile('/test/file.txt', content='content')
    assert file.deleted is False
    file.delete()
    assert file.deleted is True


def test_memory_file_repr():
    """MemoryFile has useful repr for debugging."""
    file = MemoryFile('/test/file.txt', content='content')
    assert repr(file) == snapshot("MemoryFile(path=/test/file.txt, content='...', permissions=420)")


def test_memory_file_bytes_repr():
    """MemoryFile repr shows b'...' for bytes content."""
    file = MemoryFile('/test/file.bin', content=b'\x00')
    assert repr(file) == snapshot("MemoryFile(path=/test/file.bin, content=b'...', permissions=420)")


# =============================================================================
# CallbackFile Behavior
# =============================================================================


def test_callback_file_read():
    """CallbackFile calls read callback."""
    read_calls: list[PurePosixPath] = []

    def read_fn(path: PurePosixPath) -> str:
        read_calls.append(path)
        return f'content from {path}'

    def write_fn(path: PurePosixPath, content: str | bytes) -> None:
        pass

    file = CallbackFile('/test/file.txt', read=read_fn, write=write_fn)
    fs = OSAccess([file])

    result = Sandbox('from pathlib import Path; Path("/test/file.txt").read_text()').run(os=fs)
    assert result == snapshot('content from /test/file.txt')
    assert len(read_calls) == 1


def test_callback_file_write_direct():
    """CallbackFile calls write callback via direct API."""
    written: list[tuple[PurePosixPath, Any]] = []

    def read_fn(path: PurePosixPath) -> str:
        return ''

    def write_fn(path: PurePosixPath, content: str | bytes) -> None:
        written.append((path, content))

    file = CallbackFile('/test/file.txt', read=read_fn, write=write_fn)
    fs = OSAccess([file])

    # Use direct API since write_text not implemented in Sandbox
    fs.path_write_text(P('/test/file.txt'), 'new content')
    assert len(written) == 1
    assert written[0][1] == snapshot('new content')


def test_callback_file_custom_permissions():
    """CallbackFile accepts custom permissions."""
    file = CallbackFile(
        '/test/file.txt',
        read=lambda _: '',
        write=lambda _p, _c: None,
        permissions=0o700,
    )
    assert file.permissions == snapshot(0o700)


def test_callback_file_repr():
    """CallbackFile has useful repr for debugging."""
    file = CallbackFile('/test/file.txt', read=lambda _: '', write=lambda _, __: None)
    assert 'CallbackFile(path=/test/file.txt' in repr(file)


# =============================================================================
# Custom AbstractFile Implementation
# =============================================================================


class CustomFile:
    """Minimal custom AbstractFile implementation."""

    def __init__(self, path: str, content: str) -> None:
        self.path = PurePosixPath(path)
        self.name = self.path.name
        self.permissions = 0o644
        self.deleted = False
        self.content = content

    def read_content(self) -> str:
        return self.content

    def write_content(self, content: str | bytes) -> None:
        self.content = content if isinstance(content, str) else content.decode()

    def delete(self) -> None:
        self.deleted = True


def test_custom_abstract_file():
    """Custom AbstractFile implementation works with OSAccess."""
    custom = CustomFile('/test/custom.txt', 'custom content')
    fs = OSAccess([custom])

    result = Sandbox('from pathlib import Path; Path("/test/custom.txt").read_text()').run(os=fs)
    assert result == snapshot('custom content')


def test_custom_abstract_file_mixed_with_memory_file():
    """Custom AbstractFile can be mixed with MemoryFile."""
    custom = CustomFile('/test/custom.txt', 'from custom')
    memory = MemoryFile('/test/memory.txt', content='from memory')
    fs = OSAccess([custom, memory])

    code = """
from pathlib import Path
(Path('/test/custom.txt').read_text(), Path('/test/memory.txt').read_text())
"""
    result = Sandbox(code).run(os=fs)
    assert result == snapshot(('from custom', 'from memory'))


# =============================================================================
# Direct API Test (without Sandbox)
# =============================================================================


def test_os_access_direct_api():
    """OSAccess methods can be called directly without Sandbox."""
    fs = OSAccess(
        [
            MemoryFile('/test/file.txt', content='hello'),
            MemoryFile('/test/subdir/nested.txt', content='nested'),
        ]
    )

    # Test path_exists
    assert fs.path_exists(P('/test/file.txt')) is True
    assert fs.path_exists(P('/missing')) is False

    # Test path_is_file / path_is_dir
    assert fs.path_is_file(P('/test/file.txt')) is True
    assert fs.path_is_dir(P('/test/file.txt')) is False
    assert fs.path_is_dir(P('/test/subdir')) is True
    assert fs.path_is_file(P('/test/subdir')) is False

    # Test path_read_text / path_read_bytes
    assert fs.path_read_text(P('/test/file.txt')) == 'hello'
    assert fs.path_read_bytes(P('/test/file.txt')) == b'hello'

    # Test path_stat
    stat = fs.path_stat(P('/test/file.txt'))
    assert stat.st_size == 5

    # Test path_iterdir
    contents = fs.path_iterdir(P('/test'))
    assert sorted(contents) == snapshot([PurePosixPath('/test/file.txt'), PurePosixPath('/test/subdir')])

    # Test path_absolute
    assert fs.path_absolute(P('relative')) == '/relative'
    assert fs.path_absolute(P('/absolute')) == '/absolute'


# =============================================================================
# Edge Cases
# =============================================================================


def test_root_directory():
    """Root directory '/' is handled correctly."""
    fs = OSAccess([MemoryFile('/file.txt', content='root file')])
    code = """
from pathlib import Path
(Path('/').is_dir(), sorted([str(p) for p in Path('/').iterdir()]))
"""
    result = Sandbox(code).run(os=fs)
    assert result == snapshot((True, ['/file.txt']))


def test_empty_file():
    """Empty file content is handled correctly."""
    fs = OSAccess([MemoryFile('/empty.txt', content='')])
    code = """
from pathlib import Path
(Path('/empty.txt').read_text(), Path('/empty.txt').stat().st_size)
"""
    result = Sandbox(code).run(os=fs)
    assert result == snapshot(('', 0))


def test_large_nested_path():
    """Deeply nested paths are handled correctly."""
    fs = OSAccess([MemoryFile('/a/b/c/d/e/f/g/h/i/j/file.txt', content='deep')])
    code = """
from pathlib import Path
Path('/a/b/c/d/e/f/g/h/i/j/file.txt').read_text()
"""
    result = Sandbox(code).run(os=fs)
    assert result == snapshot('deep')


def test_special_characters_in_content():
    """Special characters in file content are handled correctly."""
    content = 'line1\nline2\ttab\r\nwindows'
    fs = OSAccess([MemoryFile('/special.txt', content=content)])
    result = Sandbox('from pathlib import Path; Path("/special.txt").read_text()').run(os=fs)
    assert result == snapshot('line1\nline2\ttab\r\nwindows')



# === Basic global read/write ===
x1 = 42


def read_explicit():
    global x1
    return x1


assert read_explicit() == 42, 'explicit global read'


x2 = 1


def write_explicit():
    global x2
    x2 = 2


write_explicit()
assert x2 == 2, 'explicit global write'


x3 = 42


def read_implicit():
    return x3  # no local x3, reads global


assert read_implicit() == 42, 'implicit global read'


# === Multiple functions sharing global ===
counter1 = 0


def inc():
    global counter1
    counter1 = counter1 + 1


def get_counter():
    return counter1


inc()
inc()
assert get_counter() == 2, 'multiple functions sharing global'


# === Mutating global containers (no 'global' needed) ===
data1 = {'a': 1}


def add_dict_entry():
    data1['b'] = 2


add_dict_entry()
assert data1 == {'a': 1, 'b': 2}, 'mutate global dict'


items1 = [1, 2]


def append_list_item():
    items1.append(3)


append_list_item()
assert items1 == [1, 2, 3], 'mutate global list append'


items2 = ['a', 'c']


def insert_list_item():
    items2.insert(1, 'b')


insert_list_item()
assert items2 == ['a', 'b', 'c'], 'mutate global list insert'


items3 = []


def build_list():
    items3.append(1)
    items3.append(2)
    items3.append(3)


build_list()
assert items3 == [1, 2, 3], 'mutate global list multiple'


# === Reassigning global containers (requires 'global') ===
items4 = [1, 2]


def replace_list():
    global items4
    items4 = [3, 4, 5]


replace_list()
assert items4 == [3, 4, 5], 'reassign global list'


# === Nested functions with global ===
x4 = 1


def outer_global():
    def inner():
        global x4
        x4 = 10

    inner()


outer_global()
assert x4 == 10, 'nested inner global write'


x5 = 42


def outer_read():
    def inner():
        return x5  # reads global

    return inner()


assert outer_read() == 42, 'nested inner global read'


# === Shadowing ===
x6 = 10


def shadow_local():
    x6 = 20  # creates local (shadows global)
    return x6


assert shadow_local() == 20, 'local shadows global'


x7 = 10


def shadow_unchanged():
    x7 = 99  # local
    return x7


assert shadow_unchanged() == 99, 'shadowing returns local'
assert x7 == 10, 'global unchanged after shadowing'



# === StringIO Basic Tests ===
from io import StringIO

# Test StringIO initialization
s = StringIO()
assert s.getvalue() == '', 'Empty StringIO should have empty value'
assert s.tell() == 0, 'New StringIO position should be 0'
assert not s.closed, 'New StringIO should not be closed'
s.close()

s = StringIO('hello')
assert s.getvalue() == 'hello', 'StringIO with initial value'
s.close()

s = StringIO('')
assert s.getvalue() == '', 'StringIO with empty initial value'
s.close()

# Test StringIO with None
s = StringIO(None)
assert s.getvalue() == '', 'StringIO with None should be empty'
s.close()

# === StringIO Read Tests ===
s = StringIO('hello world')
assert s.read(2) == 'he', 'read(2) should return first 2 chars'
assert s.read(3) == 'llo', 'read(3) should return next 3 chars'
assert s.read() == ' world', 'read() should return rest'
assert s.read() == '', 'read() at EOF should return empty'
s.close()

# Test read with negative size (should read all)
s = StringIO('hello')
assert s.read(-1) == 'hello', 'read(-1) should return all'
s.close()

# === StringIO Write Tests ===
s = StringIO()
assert s.write('hello') == 5, 'write should return char count'
assert s.getvalue() == 'hello', 'getvalue after write'
s.close()

# Test write at position
s = StringIO('hello world')
s.seek(6)
s.write('Python')
assert s.getvalue() == 'hello Python', 'write should overwrite at position'
s.close()

# === StringIO Seek Tests ===
s = StringIO('hello')
assert s.tell() == 0, 'initial position is 0'
s.seek(3)
assert s.tell() == 3, 'seek(3) should set position to 3'
assert s.read() == 'lo', 'read after seek should work'
s.close()

# Test seek to end
s = StringIO('hello')
s.seek(0, 2)
assert s.tell() == 5, 'seek to end should set position to length'
s.close()

# === StringIO GetValue Tests ===
s = StringIO('test')
assert s.getvalue() == 'test', 'getvalue returns current buffer'
s.seek(2)
assert s.getvalue() == 'test', 'getvalue returns full buffer regardless of position'
s.write('X')
assert s.getvalue() == 'teXt', 'getvalue reflects writes'
s.close()

# === StringIO Close Tests ===
s = StringIO('test')
s.close()
assert s.closed, 'closed should be True after close'

try:
    s.read()
    assert False, 'read on closed should raise'
except ValueError as e:
    assert 'closed' in str(e).lower(), 'Error should mention closed'

try:
    s.write('x')
    assert False, 'write on closed should raise'
except ValueError:
    pass

# === StringIO Boolean Tests ===
s = StringIO()
assert s.readable(), 'StringIO should be readable'
assert s.writable(), 'StringIO should be writable'
assert s.seekable(), 'StringIO should be seekable'
assert not s.isatty(), 'StringIO should not be a tty'
s.close()

# === StringIO ReadLine Tests ===
s = StringIO('line1\nline2\nline3')
assert s.readline() == 'line1\n', 'readline should include newline'
assert s.readline() == 'line2\n', 'readline second call'
assert s.readline() == 'line3', 'readline last without newline'
assert s.readline() == '', 'readline at EOF'
s.close()

# === StringIO WriteLines Tests ===
s = StringIO()
s.writelines(['a', 'b', 'c'])
assert s.getvalue() == 'abc', 'writelines should concatenate'
s.close()

# === StringIO Truncate Tests ===
s = StringIO('hello world')
s.truncate(5)
assert s.getvalue() == 'hello', 'truncate should shorten buffer'
s.close()

s = StringIO('hello')
s.seek(3)
s.truncate()
assert s.getvalue() == 'hel', 'truncate to current position'
s.close()

# === StringIO Flush Test ===
s = StringIO()
assert s.flush() is None, 'flush should return None'
s.close()

# === StringIO Iteration Tests ===
s = StringIO('line1\nline2')
lines = list(s)
assert lines == ['line1\n', 'line2'], 'iteration should yield lines'
s.close()

# === StringIO Context Manager Tests ===
with StringIO() as s:
    s.write('test')
    assert s.getvalue() == 'test', 'context manager should work'
assert s.closed, 'stream should be closed after context'



import operator

# === Simple operators ===
assert operator.and_(6, 3) == 2, 'and_ should compute bitwise and for ints'
assert operator.and_(True, False) == 0, 'and_ should handle bool operands'
assert operator.or_(6, 3) == 7, 'or_ should compute bitwise or for ints'
assert operator.or_(True, False) == 1, 'or_ should handle bool operands'
assert operator.not_(False) is True, 'not_ should invert truthiness'
assert operator.inv(2) == -3, 'inv should compute bitwise invert'
assert operator.invert(2) == -3, 'invert should compute bitwise invert'

try:
    operator.matmul(1, 2)
    assert False, 'matmul should raise for integers'
except TypeError as exc:
    assert str(exc) == "unsupported operand type(s) for @: 'int' and 'int'", 'matmul error message should match'

# === Identity operators ===
obj = []
assert operator.is_(obj, obj) is True, 'is_ should be true for the same object'
assert operator.is_(obj, []) is False, 'is_ should be false for different objects'
assert operator.is_not(obj, obj) is False, 'is_not should be false for the same object'
assert operator.is_none(None) is True, 'is_none should be true for None'
assert operator.is_not_none(None) is False, 'is_not_none should be false for None'
assert operator.is_not_none(0) is True, 'is_not_none should be true for non-None'

# === In-place operators ===
assert operator.iand(6, 3) == 2, 'iand should compute bitwise and'
assert operator.ior(6, 1) == 7, 'ior should compute bitwise or'
assert operator.ixor(6, 3) == 5, 'ixor should compute bitwise xor'
assert operator.ilshift(1, 3) == 8, 'ilshift should compute left shift'
assert operator.irshift(8, 1) == 4, 'irshift should compute right shift'
assert operator.ipow(2, 3) == 8, 'ipow should compute exponentiation'

try:
    operator.imatmul(1, 2)
    assert False, 'imatmul should raise for integers'
except TypeError as exc:
    assert str(exc) == "unsupported operand type(s) for @=: 'int' and 'int'", 'imatmul error message should match'

# === Container operations ===
items = [1, 2, 3]
operator.setitem(items, 1, 99)
assert items == [1, 99, 3], 'setitem should update list element'
operator.delitem(items, 0)
assert items == [99, 3], 'delitem should remove list element'

mapping = {'a': 1}
operator.setitem(mapping, 'b', 2)
assert mapping == {'a': 1, 'b': 2}, 'setitem should insert dict entry'
operator.delitem(mapping, 'a')
assert mapping == {'b': 2}, 'delitem should remove dict entry'

assert operator.length_hint([1, 2, 3]) == 3, 'length_hint should return len for sequences'
assert operator.length_hint(10) == 0, 'length_hint should fall back to 0 when len is unavailable'

# === Sequence concatenation ===
seq = [1]
result = operator.iconcat(seq, [2, 3])
assert result == [1, 2, 3], 'iconcat should return concatenated list'
assert seq == [1, 2, 3], 'iconcat should mutate the original list'

# === Callable factories ===
get_second = operator.itemgetter(1)
assert get_second([10, 20, 30]) == 20, 'itemgetter should select the requested item'
get_pair = operator.itemgetter(0, 2)
assert get_pair([10, 20, 30]) == (10, 30), 'itemgetter should return tuples for multiple items'


class Inner:
    def __init__(self):
        self.y = 5


class Outer:
    def __init__(self):
        self.x = 3
        self.inner = Inner()


outer = Outer()
assert operator.attrgetter('x')(outer) == 3, 'attrgetter should fetch a single attribute'
assert operator.attrgetter('inner.y', 'x')(outer) == (5, 3), 'attrgetter should handle dotted names'

count_ones = operator.methodcaller('count', 1)
assert count_ones([1, 2, 1, 3]) == 2, 'methodcaller should call methods with args'
append_three = operator.methodcaller('append', 3)
values = [1, 2]
result = append_three(values)
assert result is None, 'methodcaller should return the method result'
assert values == [1, 2, 3], 'methodcaller should mutate the target when method does'

# === operator.call ===
assert operator.call(lambda a, b: a + b, 2, 3) == 5, 'operator.call should invoke callable with positional args'
assert operator.call(lambda *, x: x * 2, x=4) == 8, 'operator.call should forward keyword arguments'

# === iconcat + immutable fallback ===
immutable_result = operator.iconcat((1, 2), (3, 4))
assert immutable_result == (1, 2, 3, 4), 'iconcat should fall back to concatenation for immutable sequences'

# === length_hint default argument ===
assert operator.length_hint(10, 7) == 7, 'length_hint should return provided default when len is unavailable'


# === methodcaller kwargs support ===
class MethodCallerKwargsExample:
    def render(self, *, prefix='[', suffix=']'):
        return f'{prefix}value{suffix}'


with_kwargs = operator.methodcaller('render', prefix='<', suffix='>')
assert with_kwargs(MethodCallerKwargsExample()) == '<value>', 'methodcaller should support stored keyword args'



# call-external
# External calls in unary expressions

# Negation of external call result
result = -add_ints(3, 4)
assert result == -7, 'negation of ext call'

# Not of external call
result = not return_value(False)
assert result == True, 'not of ext call returning False'

result = not return_value(True)
assert result == False, 'not of ext call returning True'



# bool is a subclass of int in Python arithmetic

# += with bool
x = 0
x += True
assert x == 1
x += False
assert x == 1

# + with bool
assert 1 + True == 2
assert True + 1 == 2
assert True + True == 2

# - with bool
assert 5 - True == 4
assert True - 1 == 0
assert True - True == 0

# * with bool
assert 3 * True == 3
assert True * 3 == 3
assert True * True == 1

# / with bool
assert 4 / True == 4.0
assert True / 2 == 0.5

# // with bool
assert 5 // True == 5
assert True // 2 == 0

# % with bool
assert 5 % True == 0
assert True % 2 == 1

# ** with bool
assert 2 ** True == 2
assert 2 ** False == 1
assert True ** 3 == 1

# in-place ops with bool RHS
v = 5
v -= True
assert v == 4

v = 6
v *= True
assert v == 6

v = 7
v //= True
assert v == 7

v = 8
v %= True
assert v == 0

v = 2
v **= True
assert v == 2



# Verify that ModuleNotFoundError is raised when an unknown module import is actually executed
# (not guarded by TYPE_CHECKING)

condition = True
if condition:
    import nonexistent_at_runtime

"""
TRACEBACK:
Traceback (most recent call last):
  File "import__runtime_error_when_executed.py", line 6, in <module>
    import nonexistent_at_runtime
ModuleNotFoundError: No module named 'nonexistent_at_runtime'
"""



import io


# === Module-level parity ===
assert io.open is open
assert isinstance(io.DEFAULT_BUFFER_SIZE, int)
assert io.DEFAULT_BUFFER_SIZE == 131072
assert isinstance(io.TextIOBase, type)
assert isinstance(io.RawIOBase, type)
assert isinstance(io.BufferedIOBase, type)

missing = "__ouros_io_missing_file__"
try:
    io.open(missing)
    assert False, "io.open should behave like builtin open"
except FileNotFoundError as exc:
    assert missing in str(exc)


# === StringIO constructor/type validation ===
s = io.StringIO(None)
assert s.getvalue() == ""
s.close()

try:
    io.StringIO(123)
    assert False, "StringIO should reject non-str/non-None initial value"
except TypeError as exc:
    assert "initial_value must be str or None" in str(exc)

try:
    io.StringIO("a", initial_value="b")
    assert False, "duplicate initial_value should fail"
except TypeError as exc:
    msg = str(exc)
    assert ("multiple values" in msg) or ("position (1)" in msg)

for newline in (None, "", "\n", "\r", "\r\n"):
    s = io.StringIO("x", newline=newline)
    assert s.getvalue() == "x"
    s.close()

try:
    io.StringIO("", newline=1)
    assert False, "newline must be str or None"
except TypeError as exc:
    assert "newline must be str or None" in str(exc)

try:
    io.StringIO("", newline="bad")
    assert False, "invalid newline must raise ValueError"
except ValueError as exc:
    assert "illegal newline value" in str(exc)


# === StringIO semantics ===
s = io.StringIO("abc")
s.seek(5)
assert s.write("Z") == 1
assert s.tell() == 6
assert s.getvalue() == "abc\0\0Z"
s.close()

s = io.StringIO("abc")
s.seek(5)
assert s.write("") == 0
assert s.tell() == 5
assert s.getvalue() == "abc"
s.close()

s = io.StringIO("caf\u00e9")
assert s.read(3) == "caf"
assert s.read() == "\u00e9"
s.seek(3)
assert s.write("X") == 1
assert s.getvalue() == "cafX"
s.close()

s = io.StringIO("a\nb\nccc\n")
assert s.readlines(1) == ["a\n"]
s.seek(0)
assert s.readlines(2) == ["a\n", "b\n"]
s.seek(0)
assert s.readlines(4) == ["a\n", "b\n", "ccc\n"]
s.close()

try:
    io.StringIO().write(1)
    assert False, "write() must only accept str"
except TypeError as exc:
    assert "string argument expected" in str(exc)

try:
    io.StringIO().writelines([1])
    assert False, "writelines() must only accept str elements"
except TypeError as exc:
    assert "string argument expected" in str(exc)

s = io.StringIO("line1\nline2")
assert s.__iter__() is s
assert next(s) == "line1\n"
assert next(s) == "line2"
try:
    next(s)
    assert False
except StopIteration:
    pass

s = io.StringIO("x")
assert s.__exit__(None, None, None) is None
assert s.closed


# === BytesIO constructor/type validation ===
b = io.BytesIO(bytearray(b"abc"))
assert b.getvalue() == b"abc"
b.close()

try:
    io.BytesIO(123)
    assert False, "BytesIO should reject non-bytes-like initial value"
except TypeError as exc:
    assert "bytes-like object" in str(exc)

try:
    io.BytesIO(b"a", initial_bytes=b"b")
    assert False, "duplicate initial_bytes should fail"
except TypeError as exc:
    msg = str(exc)
    assert ("multiple values" in msg) or ("at most 1 argument" in msg)


# === BytesIO semantics ===
b = io.BytesIO(b"abc")
b.seek(5)
assert b.write(b"Z") == 1
assert b.tell() == 6
assert b.getvalue() == b"abc\x00\x00Z"
b.close()

b = io.BytesIO(b"abc")
b.seek(5)
assert b.write(b"") == 0
assert b.tell() == 5
assert b.getvalue() == b"abc"
b.close()

b = io.BytesIO(b"a\nb\nccc\n")
assert b.readlines(1) == [b"a\n"]
b.seek(0)
assert b.readlines(2) == [b"a\n"]
b.seek(0)
assert b.readlines(4) == [b"a\n", b"b\n"]
b.close()

b = io.BytesIO()
assert b.write(bytearray(b"xy")) == 2
assert b.getvalue() == b"xy"
b.close()

try:
    io.BytesIO().write(1)
    assert False, "BytesIO.write() must only accept bytes-like values"
except TypeError as exc:
    assert "bytes-like object" in str(exc)

try:
    io.BytesIO().writelines([1])
    assert False, "BytesIO.writelines() must only accept bytes-like elements"
except TypeError as exc:
    assert "bytes-like object" in str(exc)

b = io.BytesIO(b"line1\nline2")
assert b.__iter__() is b
assert next(b) == b"line1\n"
assert next(b) == b"line2"
try:
    next(b)
    assert False
except StopIteration:
    pass

b = io.BytesIO(b"abc")
view = b.getbuffer()
assert bytes(view) == b"abc"
try:
    b.close()
    assert False, "close should fail while exported buffer is live"
except BufferError:
    pass
del view
b.close()
assert b.closed

b = io.BytesIO(b"x")
assert b.__exit__(None, None, None) is None
assert b.closed



from dataclasses import (
    FrozenInstanceError,
    asdict,
    astuple,
    dataclass,
    fields,
    is_dataclass,
)
from typing import NoReturn

import pytest
from inline_snapshot import snapshot

import ouros


@dataclass
class Person:
    name: str
    age: int


def test_dataclass_input():
    """Dataclass instances are converted and returned as OurosDataclass."""

    m = ouros.Sandbox('x', inputs=['x'])
    m.register_dataclass(Person)
    result = m.run(inputs={'x': Person(name='Alice', age=30)})
    assert result.name == snapshot('Alice')
    assert result.age == snapshot(30)
    assert is_dataclass(result)
    assert isinstance(result, Person)
    assert asdict(result) == snapshot({'name': 'Alice', 'age': 30})
    assert repr(result) == snapshot("Person(name='Alice', age=30)")


def test_dataclass_unknown():
    """Dataclass instances are converted and returned as OurosDataclass."""

    m = ouros.Sandbox('x', inputs=['x'])
    result = m.run(inputs={'x': Person(name='Alice', age=30)})
    assert result.name == snapshot('Alice')
    assert result.age == snapshot(30)
    assert is_dataclass(result)
    assert not isinstance(result, Person)
    assert asdict(result) == snapshot({'name': 'Alice', 'age': 30})  # pyright: ignore[reportArgumentType]
    assert repr(result) == snapshot("<Unknown Dataclass Person(name='Alice', age=30)>")


@dataclass(frozen=True)
class Point:
    x: int
    y: int


def test_dataclass_frozen():
    """Frozen dataclasses are converted like regular dataclasses."""

    m = ouros.Sandbox('p', inputs=['p'], dataclass_registry=[Point])
    result = m.run(inputs={'p': Point(x=10, y=20)})
    assert isinstance(result, Point)
    assert result.x == snapshot(10)
    assert result.y == snapshot(20)
    assert repr(result) == snapshot('Point(x=10, y=20)')


@dataclass
class Address:
    city: str
    zip_code: str


@dataclass
class PersonAddress:
    name: str
    address: Address


def test_dataclass_nested():
    """Nested dataclasses are recursively converted."""

    m = ouros.Sandbox('x', inputs=['x'])
    m.register_dataclass(Address)
    m.register_dataclass(PersonAddress)
    result = m.run(inputs={'x': PersonAddress(name='Bob', address=Address(city='NYC', zip_code='10001'))})
    assert isinstance(result, PersonAddress)
    assert result.name == snapshot('Bob')
    assert isinstance(result.address, Address)
    assert result.address.city == snapshot('NYC')
    assert result.address.zip_code == snapshot('10001')


def test_dataclass_nested_unknown():
    m = ouros.Sandbox('x', inputs=['x'])
    result = m.run(inputs={'x': PersonAddress(name='Bob', address=Address(city='NYC', zip_code='10001'))})
    assert not isinstance(result, PersonAddress)
    assert result.name == snapshot('Bob')
    assert not isinstance(result.address, Address)
    assert result.address.city == snapshot('NYC')
    assert result.address.zip_code == snapshot('10001')


def test_dataclass_with_list_field():
    """Dataclasses with list fields are properly converted."""

    @dataclass
    class Container:
        items: list[int]

    m = ouros.Sandbox('x', inputs=['x'])
    m.register_dataclass(Container)
    result = m.run(inputs={'x': Container(items=[1, 2, 3])})
    assert result.items == snapshot([1, 2, 3])


def test_dataclass_with_dict_field():
    """Dataclasses with dict fields are properly converted."""

    @dataclass
    class Config:
        settings: dict[str, int]

    m = ouros.Sandbox('x', inputs=['x'])
    m.register_dataclass(Config)
    result = m.run(inputs={'x': Config(settings={'a': 1, 'b': 2})})
    assert result.settings == snapshot({'a': 1, 'b': 2})


def test_dataclass_empty():
    """Empty dataclass (no fields) has empty repr."""

    @dataclass
    class Empty:
        pass

    m = ouros.Sandbox('x', inputs=['x'])
    m.register_dataclass(Empty)
    result = m.run(inputs={'x': Empty()})
    assert repr(result) == snapshot('test_dataclass_empty.<locals>.Empty()')


def test_dataclass_type_raises():
    """Dataclass type (not instance) should raise TypeError."""

    @dataclass
    class MyClass:
        value: int

    m = ouros.Sandbox('x', inputs=['x'])
    m.register_dataclass(MyClass)
    with pytest.raises(TypeError, match='Cannot convert type to Sandbox value'):
        m.run(inputs={'x': MyClass})


# === Field access ===


def test_dataclass_field_access():
    """Access individual fields of a dataclass."""

    @dataclass
    class Person:
        name: str
        age: int

    m = ouros.Sandbox('x.name', inputs=['x'])
    assert m.run(inputs={'x': Person(name='Alice', age=30)}) == snapshot('Alice')

    m = ouros.Sandbox('x.age', inputs=['x'])
    assert m.run(inputs={'x': Person(name='Alice', age=30)}) == snapshot(30)


def test_dataclass_field_access_nested():
    """Access fields of nested dataclasses."""

    m = ouros.Sandbox('x.address.city', inputs=['x'])
    result = m.run(inputs={'x': PersonAddress(name='Bob', address=Address(city='NYC', zip_code='10001'))})
    assert result == snapshot('NYC')


def test_dataclass_field_in_expression():
    """Use dataclass fields in expressions."""

    @dataclass
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('p.x + p.y', inputs=['p'])
    assert m.run(inputs={'p': Point(x=10, y=20)}) == snapshot(30)


def test_dataclass_field_access_missing():
    """Accessing a non-existent field raises AttributeError."""

    @dataclass
    class Person:
        name: str

    m = ouros.Sandbox('x.age', inputs=['x'])
    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        m.run(inputs={'x': Person(name='Alice')})
    assert isinstance(exc_info.value.exception(), AttributeError)


# === Repr ===


def test_dataclass_repr():
    """Repr of dataclass shows ClassName(field=value, ...) format."""

    @dataclass
    class Person:
        name: str
        age: int

    m = ouros.Sandbox('repr(x)', inputs=['x'])
    assert m.run(inputs={'x': Person(name='Alice', age=30)}) == snapshot("Person(name='Alice', age=30)")


def test_dataclass_repr_frozen():
    """Repr of frozen dataclass shows same format."""

    @dataclass(frozen=True)
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('repr(p)', inputs=['p'])
    assert m.run(inputs={'p': Point(x=10, y=20)}) == snapshot('Point(x=10, y=20)')


def test_dataclass_repr_nested():
    """Repr of nested dataclass shows nested repr."""

    @dataclass
    class Inner:
        value: int

    @dataclass
    class Outer:
        inner: Inner

    m = ouros.Sandbox('repr(x)', inputs=['x'])
    assert m.run(inputs={'x': Outer(inner=Inner(value=42))}) == snapshot('Outer(inner=Inner(value=42))')


def test_dataclass_repr_empty():
    """Repr of empty dataclass shows ClassName()."""

    @dataclass
    class Empty:
        pass

    m = ouros.Sandbox('repr(x)', inputs=['x'])
    m.register_dataclass(Empty)
    assert m.run(inputs={'x': Empty()}) == snapshot('Empty()')


# === Setattr ===


def test_dataclass_setattr_mutable_unknown():
    """Setting attributes on mutable dataclass works."""

    @dataclass
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('p', inputs=['p'])
    result = m.run(inputs={'p': Point(x=10, y=20)})

    # Modify existing field
    result.x = 100
    assert result.x == snapshot(100)
    assert repr(result) == snapshot('<Unknown Dataclass Point(x=100, y=20)>')

    # Add new attribute (not in repr since not a declared field)
    result.z = 30
    assert result.z == snapshot(30)
    assert repr(result) == snapshot('<Unknown Dataclass Point(x=100, y=20)>')


def test_dataclass_setattr_frozen():
    """Setting attributes on frozen dataclass raises FrozenInstanceError."""

    @dataclass(frozen=True)
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('p', inputs=['p'])
    result = m.run(inputs={'p': Point(x=10, y=20)})

    # FrozenInstanceError is raised (which is a subclass of AttributeError)
    with pytest.raises(FrozenInstanceError, match="cannot assign to field 'x'"):
        result.x = 100

    with pytest.raises(FrozenInstanceError, match="cannot assign to field 'z'"):
        result.z = 30


def test_frozen_instance_error_is_attribute_error():
    """FrozenInstanceError can be caught as AttributeError."""

    @dataclass(frozen=True)
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('p', inputs=['p'])
    result = m.run(inputs={'p': Point(x=10, y=20)})

    # Can catch with AttributeError (parent class)
    with pytest.raises(AttributeError):
        result.x = 100

    # Verify it's actually FrozenInstanceError
    try:
        result.y = 200
    except AttributeError as e:
        assert isinstance(e, FrozenInstanceError)


def test_frozen_instance_error_message():
    """FrozenInstanceError has correct message format."""

    @dataclass(frozen=True)
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('p', inputs=['p'])
    result = m.run(inputs={'p': Point(x=10, y=20)})

    with pytest.raises(FrozenInstanceError) as exc_info:
        result.x = 100
    assert exc_info.value.args[0] == snapshot("cannot assign to field 'x'")


def test_frozen_instance_error_from_ouros_code():
    """FrozenInstanceError raised by Sandbox code is properly converted."""

    @dataclass(frozen=True)
    class Point:
        x: int
        y: int

    # Sandbox code that tries to modify a frozen dataclass
    code = """
p.x = 100
"""
    m = ouros.Sandbox(code, inputs=['p'])

    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        m.run(inputs={'p': Point(x=10, y=20)})
    inner = exc_info.value.exception()
    assert isinstance(inner, FrozenInstanceError)
    assert inner.args[0] == snapshot("cannot assign to field 'x'")


def test_frozen_instance_error_from_ouros_caught_as_attribute_error():
    """FrozenInstanceError from Sandbox can be caught as AttributeError."""

    @dataclass(frozen=True)
    class Point:
        x: int
        y: int

    code = 'p.x = 100'
    m = ouros.Sandbox(code, inputs=['p'])

    # Wrapped in SandboxRuntimeError, but inner exception is FrozenInstanceError
    # which is a subclass of AttributeError
    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        m.run(inputs={'p': Point(x=10, y=20)})
    inner = exc_info.value.exception()
    assert isinstance(inner, AttributeError)
    assert isinstance(inner, FrozenInstanceError)


def test_frozen_instance_error_from_external_function():
    """FrozenInstanceError from external function is properly converted."""
    code = """
try:
    fail()
except FrozenInstanceError:
    caught = 'frozen'
except AttributeError:
    caught = 'attr'
caught
"""
    m = ouros.Sandbox(code, external_functions=['fail'])

    def fail() -> NoReturn:
        raise FrozenInstanceError('cannot assign to field')

    # Sandbox should catch it as FrozenInstanceError specifically
    result = m.run(external_functions={'fail': fail})
    assert result == snapshot('frozen')


def test_frozen_instance_error_from_external_function_propagates():
    """FrozenInstanceError from external function propagates to Python."""
    m = ouros.Sandbox('fail()', external_functions=['fail'])

    def fail() -> NoReturn:
        raise FrozenInstanceError('test frozen error')

    with pytest.raises(ouros.SandboxRuntimeError) as exc_info:
        m.run(external_functions={'fail': fail})
    inner = exc_info.value.exception()
    assert isinstance(inner, FrozenInstanceError)
    assert inner.args[0] == snapshot('test frozen error')


# === Equality ===


def test_dataclass_equality_same():
    """Equal dataclasses compare equal."""

    @dataclass
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('(a, b)', inputs=['a', 'b'])
    a, b = m.run(inputs={'a': Point(x=10, y=20), 'b': Point(x=10, y=20)})
    assert a == b


def test_dataclass_equality_different_values():
    """Dataclasses with different values compare not equal."""

    @dataclass
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('(a, b)', inputs=['a', 'b'])
    a, b = m.run(inputs={'a': Point(x=10, y=20), 'b': Point(x=10, y=30)})
    assert a != b


def test_dataclass_equality_different_types():
    """Dataclasses of different types compare not equal."""

    @dataclass
    class Point:
        x: int
        y: int

    @dataclass
    class Vector:
        x: int
        y: int

    m = ouros.Sandbox('(a, b)', inputs=['a', 'b'])
    a, b = m.run(inputs={'a': Point(x=10, y=20), 'b': Vector(x=10, y=20)})
    assert a != b


def test_dataclass_equality_with_other_type():
    """Dataclass compared to non-dataclass returns False."""

    @dataclass
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('p', inputs=['p'])
    result = m.run(inputs={'p': Point(x=10, y=20)})
    assert result != {'x': 10, 'y': 20}
    assert result != (10, 20)
    assert result != 'Point(x=10, y=20)'


# === Hashing ===


def test_dataclass_hash_frozen():
    """Frozen dataclasses are hashable."""

    @dataclass(frozen=True)
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('p', inputs=['p'])
    result = m.run(inputs={'p': Point(x=10, y=20)})

    h = hash(result)
    assert isinstance(h, int)
    # Hash is consistent
    assert hash(result) == h


def test_dataclass_hash_frozen_equal_values():
    """Equal frozen dataclasses have equal hashes."""

    @dataclass(frozen=True)
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('(a, b)', inputs=['a', 'b'])
    a, b = m.run(inputs={'a': Point(x=10, y=20), 'b': Point(x=10, y=20)})

    assert hash(a) == hash(b)


def test_dataclass_hash_mutable_raises():
    """Mutable dataclasses are not hashable."""

    @dataclass
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('p', inputs=['p'])
    result = m.run(inputs={'p': Point(x=10, y=20)})

    with pytest.raises(TypeError, match="unhashable type: 'UnknownDataclass'"):
        hash(result)


def test_dataclass_hash_in_set():
    """Frozen dataclasses can be used in sets."""

    @dataclass(frozen=True)
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('(a, b, c)', inputs=['a', 'b', 'c'])
    a, b, c = m.run(
        inputs={
            'a': Point(x=10, y=20),
            'b': Point(x=10, y=20),  # duplicate
            'c': Point(x=30, y=40),
        }
    )

    s = {a, b, c}
    assert len(s) == snapshot(2)


def test_dataclass_hash_as_dict_key():
    """Frozen dataclasses can be used as dict keys."""

    @dataclass(frozen=True)
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('(a, b)', inputs=['a', 'b'])
    a, b = m.run(inputs={'a': Point(x=10, y=20), 'b': Point(x=10, y=20)})

    d = {a: 'first'}
    assert d[b] == snapshot('first')


# === dataclasses module compatibility ===


def test_dataclass_is_dataclass():
    """is_dataclass() returns True for returned dataclasses."""

    @dataclass
    class Person:
        name: str
        age: int

    m = ouros.Sandbox('x', inputs=['x'])
    result = m.run(inputs={'x': Person(name='Alice', age=30)})
    assert is_dataclass(result) is True


def test_dataclass_fields():
    """fields() returns Field objects for returned dataclasses."""

    @dataclass
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('p', inputs=['p'])
    result = m.run(inputs={'p': Point(x=10, y=20)})

    fs = fields(result)
    assert len(fs) == snapshot(2)
    assert fs[0].name == snapshot('x')
    assert fs[1].name == snapshot('y')
    # Type is inferred from value
    assert fs[0].type is int
    assert fs[1].type is int


def test_dataclass_fields_string():
    """fields() returns correct type for string fields."""

    @dataclass
    class Person:
        name: str

    m = ouros.Sandbox('p', inputs=['p'])
    result = m.run(inputs={'p': Person(name='Alice')})

    fs = fields(result)
    assert fs[0].name == snapshot('name')
    assert fs[0].type is str


def test_dataclass_asdict():
    """asdict() converts returned dataclass to dict."""

    @dataclass
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('p', inputs=['p'])
    result = m.run(inputs={'p': Point(x=10, y=20)})

    d = asdict(result)
    assert d == snapshot({'x': 10, 'y': 20})


def test_dataclass_asdict_nested():
    """asdict() recursively converts nested dataclasses."""

    @dataclass
    class Inner:
        value: int

    @dataclass
    class Outer:
        inner: Inner

    m = ouros.Sandbox('x', inputs=['x'])
    result = m.run(inputs={'x': Outer(inner=Inner(value=42))})

    d = asdict(result)
    assert d == snapshot({'inner': {'value': 42}})


def test_dataclass_astuple():
    """astuple() converts returned dataclass to tuple."""

    @dataclass
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('p', inputs=['p'])
    result = m.run(inputs={'p': Point(x=10, y=20)})

    t = astuple(result)
    assert t == snapshot((10, 20))


def test_dataclass_dataclass_fields_attr():
    """__dataclass_fields__ attribute is accessible."""

    @dataclass
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('p', inputs=['p'])
    result = m.run(inputs={'p': Point(x=10, y=20)})

    df = result.__dataclass_fields__
    assert 'x' in df
    assert 'y' in df
    assert df['x'].name == snapshot('x')
    assert df['y'].name == snapshot('y')


def test_dataclass_params_frozen():
    """__dataclass_params__.frozen reflects frozen status."""

    @dataclass(frozen=True)
    class FrozenPoint:
        x: int
        y: int

    @dataclass
    class MutablePoint:
        x: int
        y: int

    m = ouros.Sandbox('(f, m)', inputs=['f', 'm'])
    frozen, mutable = m.run(inputs={'f': FrozenPoint(x=1, y=2), 'm': MutablePoint(x=3, y=4)})

    assert frozen.__dataclass_params__.frozen is True
    assert mutable.__dataclass_params__.frozen is False


def test_dataclass_params_attributes():
    """__dataclass_params__ has expected attributes."""

    @dataclass
    class Point:
        x: int
        y: int

    m = ouros.Sandbox('p', inputs=['p'])
    result = m.run(inputs={'p': Point(x=10, y=20)})

    params = result.__dataclass_params__
    assert params.init is True
    assert params.repr is True
    assert params.eq is True
    assert params.order is False
    assert params.frozen is False


def test_repeat_dataclass_name():
    """Two classes with the same name are distinguished because we use id, not name."""

    def create_point():
        @dataclass
        class Point:
            x: int
            y: int

        return Point

    point_cls2 = create_point()
    m = ouros.Sandbox('a, b', inputs=['a', 'b'], dataclass_registry=[Point, point_cls2])
    a, b = m.run(inputs={'a': Point(x=10, y=20), 'b': point_cls2(x=30, y=40)})
    assert isinstance(a, Point)
    assert isinstance(b, point_cls2)



# run-async
# Async function with arguments


async def add(a, b):
    return a + b


result = await add(10, 20)  # pyright: ignore
assert result == 30, 'async function should handle arguments'

# With keyword arguments
result2 = await add(a=5, b=15)  # pyright: ignore
assert result2 == 20, 'async function should handle keyword arguments'



import enum

# === Sentinel imports ===
# These should be truthy sentinel values (not None)
assert enum.Enum is not None, 'Enum should not be None'
assert enum.IntEnum is not None, 'IntEnum should not be None'
assert enum.Flag is not None, 'Flag should not be None'
assert enum.IntFlag is not None, 'IntFlag should not be None'
assert enum.StrEnum is not None, 'StrEnum should not be None'

# === Existence checks ===
assert enum.unique is not None, 'unique should exist'
assert enum.member is not None, 'member should exist'
assert enum.nonmember is not None, 'nonmember should exist'
assert str(enum.CONFORM) == 'conform', 'CONFORM should stringify to conform'
assert str(enum.EJECT) == 'eject', 'EJECT should stringify to eject'
assert str(enum.KEEP) == 'keep', 'KEEP should stringify to keep'
assert str(enum.STRICT) == 'strict', 'STRICT should stringify to strict'


# === enum.unique ===
class MemberRef:
    def __init__(self, name):
        self.name = name


class UniqueLike:
    _a = MemberRef('A')
    _b = MemberRef('B')
    __members__ = {'A': _a, 'B': _b}


assert enum.unique(UniqueLike) is UniqueLike, 'unique returns class when no duplicates'


class AliasLike:
    _a = MemberRef('A')
    __members__ = {'A': _a, 'B': _a}


got_value_error = False
try:
    enum.unique(AliasLike)
except ValueError:
    got_value_error = True
assert got_value_error, 'unique raises ValueError for duplicate aliases'

# === from enum import ... ===
from enum import Enum, Flag, IntEnum, IntFlag, StrEnum

assert Enum is not None, 'from enum import Enum should work'
assert IntEnum is not None, 'from enum import IntEnum should work'
assert Flag is not None, 'from enum import Flag should work'
assert IntFlag is not None, 'from enum import IntFlag should work'
assert StrEnum is not None, 'from enum import StrEnum should work'

# === enum.auto ===
auto_a = enum.auto()
auto_b = enum.auto()
assert auto_a != auto_b, 'separate auto() calls should produce distinct values'


# === Flag class syntax + bitwise operators ===
class Permission(enum.Flag):
    READ = enum.auto()
    WRITE = enum.auto()
    EXECUTE = enum.auto()


assert Permission.READ != Permission.WRITE, 'Flag members should be distinct'
assert Permission.WRITE != Permission.EXECUTE, 'Flag members should be distinct'

permission_rw = Permission.READ | Permission.WRITE
assert (permission_rw & Permission.READ) == Permission.READ, 'Flag should support bitwise AND'
assert ((Permission.READ ^ Permission.WRITE) ^ Permission.WRITE) == Permission.READ, 'Flag should support bitwise XOR'
assert (~Permission.READ) != Permission.READ, 'Flag should support bitwise invert'


# === IntFlag class syntax + int behavior ===
class PermissionInt(enum.IntFlag):
    READ = enum.auto()
    WRITE = enum.auto()
    EXECUTE = enum.auto()


assert isinstance(PermissionInt.READ, int), 'IntFlag members should behave as ints'
permission_int_rw = PermissionInt.READ | PermissionInt.WRITE
assert (permission_int_rw & PermissionInt.READ) == PermissionInt.READ, 'IntFlag should support bitwise AND'
assert ((PermissionInt.READ ^ PermissionInt.WRITE) ^ PermissionInt.WRITE) == PermissionInt.READ, (
    'IntFlag should support bitwise XOR'
)
assert (~PermissionInt.READ) != PermissionInt.READ, 'IntFlag should support bitwise invert'



[1, 2, 3].pop(10)
"""
TRACEBACK:
Traceback (most recent call last):
  File "list__pop_out_of_range.py", line 1, in <module>
    [1, 2, 3].pop(10)
    ~~~~~~~~~~~~~~~~~
IndexError: pop index out of range
"""



# === Basic walrus operator ===
# Simple assignment expression
assert (x := 5) == 5, 'walrus returns value'
assert x == 5, 'walrus assigns to variable'

# Walrus in parentheses
y = (z := 10)
assert y == 10, 'walrus value can be assigned'
assert z == 10, 'walrus target is assigned'

# simple if
x = None
answer = 'unset'
if y := x:
    answer = f'x is {y}'

assert answer == 'unset'

x = 123
if y := x:
    answer = f'x is {y}'

assert answer == 'x is 123'
x = 0
if y := x:
    answer = f'x is {y}'
else:
    answer = 'x is unset'

assert answer == 'x is unset'

# === Walrus in if conditions ===
if (a := 3) > 0:
    assert a == 3, 'walrus in if test'
else:
    assert False, 'should not reach else'

# With falsy value
if b := 0:
    assert False, 'should not reach truthy branch'
else:
    assert b == 0, 'walrus assigns even when falsy'

# === Walrus in while loops ===
counter = 0
result = []
while (n := counter) < 3:
    result.append(n)
    counter += 1
assert result == [0, 1, 2], 'walrus in while condition'
assert n == 3, 'walrus value persists after while'

# === Nested walrus ===
# Inner walrus assigned first, then outer
assert (outer := (inner := 7) + 1) == 8, 'nested walrus returns correct value'
assert inner == 7, 'inner walrus assigned'
assert outer == 8, 'outer walrus assigned'

# === Walrus in list literals ===
items = [(v := 1), v + 1, v + 2]
assert items == [1, 2, 3], 'walrus in list literal'
assert v == 1, 'walrus variable accessible after list'

# === Walrus in ternary expressions ===
result = (t := 5) if True else 0
assert result == 5, 'walrus in ternary truthy branch'
assert t == 5, 'walrus assigned in ternary'

result2 = 0 if False else (f := 6)
assert result2 == 6, 'walrus in ternary falsy branch'
assert f == 6, 'walrus assigned in falsy branch'

# === Walrus in dict/set literals ===
d = {(k := 'key'): (val := 42)}
assert d == {'key': 42}, 'walrus in dict literal'
assert k == 'key', 'walrus key assigned'
assert val == 42, 'walrus value assigned'

s = {(s1 := 1), (s2 := 2)}
assert s == {1, 2}, 'walrus in set literal'
assert s1 == 1, 'walrus in set element 1'
assert s2 == 2, 'walrus in set element 2'

# === Walrus in subscript expressions ===
arr = [10, 20, 30]
value = arr[(idx := 1)]
assert value == 20, 'walrus in subscript index'
assert idx == 1, 'walrus index assigned'


# === Walrus in function calls ===
def identity(x):
    return x


result = identity((arg := 99))
assert result == 99, 'walrus in function argument'
assert arg == 99, 'walrus arg assigned'

# === Walrus with comparison operators ===
assert (cmp := 10) > 5, 'walrus in comparison'
assert cmp == 10, 'walrus assigned in comparison'

# === Walrus in chained comparisons ===
# Note: Chained comparisons like `0 < (mid := 5) < 10` are not yet supported
# Testing a simpler comparison chain
mid = (chain := 5)
assert 0 < chain and chain < 10, 'walrus result used in comparison chain'
assert mid == 5, 'walrus assigned correctly'

# === Walrus in boolean expressions ===
# Short-circuit with and
result = (first := 1) and (second := 2)
assert result == 2, 'walrus in and expression'
assert first == 1, 'first walrus assigned'
assert second == 2, 'second walrus assigned (and evaluated)'

# Short-circuit with or (second not evaluated)
result = (or_first := 1) or (or_skip := 999)
assert result == 1, 'walrus in or expression (short-circuit)'
assert or_first == 1, 'or first walrus assigned'

# === Walrus with operations ===
assert (op := 3) + 2 == 5, 'walrus with addition'
assert op == 3, 'walrus assigned before operation'

# === Walrus in f-strings ===
msg = f'{(fvar := "hello")} world'
assert msg == 'hello world', 'walrus in f-string'
assert fvar == 'hello', 'walrus assigned in f-string'

# === Walrus with global ===
global_var = None


def set_global():
    global global_var
    return (global_var := 'set')


result = set_global()
assert result == 'set', 'walrus with global returns value'
assert global_var == 'set', 'global var assigned via walrus'


# === Walrus creates local in function scope ===
def func_scope():
    if local := 42:
        pass
    return local


assert func_scope() == 42, 'walrus creates local in function'

# === Walrus in list comprehension element (leaks to enclosing scope) ===
# Per PEP 572, walrus in comprehension assigns to enclosing scope
# Note: walrus in comprehension iterable is not allowed, but in element/condition it is
result = [(leak := x) for x in range(3)]
assert result == [0, 1, 2], 'walrus in comprehension element'
assert leak == 2, 'walrus in comprehension leaks to enclosing scope'

# === Walrus in comprehension condition ===
result = [x for x in range(5) if (limit := 3) and x < limit]
assert result == [0, 1, 2], 'walrus in comprehension condition'
assert limit == 3, 'walrus from comprehension condition accessible'

# === Multiple walrus in same expression ===
result = (m1 := 1) + (m2 := 2) + (m3 := 3)
assert result == 6, 'multiple walrus in expression'
assert m1 == 1, 'first multi-walrus'
assert m2 == 2, 'second multi-walrus'
assert m3 == 3, 'third multi-walrus'

# === Walrus in tuple ===
tup = ((t1 := 'a'), (t2 := 'b'))
assert tup == ('a', 'b'), 'walrus in tuple'
assert t1 == 'a', 'first tuple walrus'
assert t2 == 'b', 'second tuple walrus'



slice(1, 2, 3, 4)
"""
TRACEBACK:
Traceback (most recent call last):
  File "slice__too_many_args.py", line 1, in <module>
    slice(1, 2, 3, 4)
    ~~~~~~~~~~~~~~~~~
TypeError: slice expected at most 3 arguments, got 4
"""



# Tests for sys module import

import sys

# === sys.version ===
# Check that version is a non-empty string (exact value differs between interpreters)
assert isinstance(sys.version, str), 'version should be a string'
assert len(sys.version) > 0, 'version should be non-empty'

# === sys.version_info ===
# Test index access returns integers for first 3 elements
assert isinstance(sys.version_info[0], int), 'major version should be int'
assert isinstance(sys.version_info[1], int), 'minor version should be int'
assert isinstance(sys.version_info[2], int), 'micro version should be int'
assert isinstance(sys.version_info[3], str), 'releaselevel should be str'
assert isinstance(sys.version_info[4], int), 'serial should be int'

# Test negative indexing
assert sys.version_info[-1] == sys.version_info[4], 'negative index -1 should equal index 4'
assert sys.version_info[-2] == sys.version_info[3], 'negative index -2 should equal index 3'
assert sys.version_info[-5] == sys.version_info[0], 'negative index -5 should equal index 0'

# Test named attribute access matches index access
assert sys.version_info.major == sys.version_info[0], 'major attr should equal index 0'
assert sys.version_info.minor == sys.version_info[1], 'minor attr should equal index 1'
assert sys.version_info.micro == sys.version_info[2], 'micro attr should equal index 2'
assert sys.version_info.releaselevel == sys.version_info[3], 'releaselevel attr should equal index 3'
assert sys.version_info.serial == sys.version_info[4], 'serial attr should equal index 4'

# Test len
assert len(sys.version_info) == 5, 'version_info should have 5 elements'

# Test tuple equality (works after fixing NamedTuple equality)
v = sys.version_info
assert (v[0], v[1]) == (v.major, v.minor), 'tuple of indices should equal tuple of attrs'
assert v.major == v[0], 'major attr should equal index 0'
assert v.minor == v[1], 'minor attr should equal index 1'

# === sys.platform ===
# Check that platform is a non-empty string (exact value differs between interpreters)
assert isinstance(sys.platform, str), 'platform should be a string'
assert len(sys.platform) > 0, 'platform should be non-empty'

# === sys.stdout and sys.stderr ===
# These should exist - we test by accessing them (will fail if not present)
stdout = sys.stdout
stderr = sys.stderr



# === Construction ===
s = set()
assert len(s) == 0, 'empty set len'
assert s == set(), 'empty set equality'

s = set([1, 2, 3])
assert len(s) == 3, 'set from list len'

# === Basic Methods ===
s = set()
s.add(1)
s.add(2)
s.add(1)  # duplicate
assert len(s) == 2, 'add with duplicate'

# === Discard and Remove ===
s = set([1, 2, 3])
s.discard(2)
assert len(s) == 2, 'discard existing'
s.discard(99)  # should not raise
assert len(s) == 2, 'discard non-existing'

# === Pop ===
s = set([1])
v = s.pop()
assert v == 1, 'pop returns element'
assert len(s) == 0, 'pop removes element'

# === Clear ===
s = set([1, 2, 3])
s.clear()
assert len(s) == 0, 'clear'

# === Copy ===
s = set([1, 2, 3])
s2 = s.copy()
assert s == s2, 'copy equality'
s.add(4)
assert s != s2, 'copy is independent'

# === Update ===
s = set([1, 2])
s.update([2, 3, 4])
assert len(s) == 4, 'update with list'
s = set([1, 2])
s.update()
assert s == set([1, 2]), 'update with no args is no-op'

# === Mutating set algebra methods ===
s = set([1, 2, 3])
s.intersection_update([2, 3, 4])
assert s == set([2, 3]), 'intersection_update basic'

s = set([1, 2, 3, 4])
s.difference_update([2, 4])
assert s == set([1, 3]), 'difference_update basic'

s = set([1, 2, 3])
s.symmetric_difference_update(set([3, 4]))
assert s == set([1, 2, 4]), 'symmetric_difference_update basic'

# === Union ===
s1 = set([1, 2])
s2 = set([2, 3])
u = s1.union(s2)
assert len(u) == 3, 'union len'

# === Intersection ===
s1 = set([1, 2, 3])
s2 = set([2, 3, 4])
i = s1.intersection(s2)
assert len(i) == 2, 'intersection len'

# === Difference ===
s1 = set([1, 2, 3])
s2 = set([2, 3, 4])
d = s1.difference(s2)
assert len(d) == 1, 'difference len'

# === Symmetric Difference ===
s1 = set([1, 2, 3])
s2 = set([2, 3, 4])
sd = s1.symmetric_difference(s2)
assert len(sd) == 2, 'symmetric_difference len'

# === Issubset ===
s1 = set([1, 2])
s2 = set([1, 2, 3])
assert s1.issubset(s2) == True, 'issubset true'
assert s2.issubset(s1) == False, 'issubset false'

# === Issuperset ===
s1 = set([1, 2, 3])
s2 = set([1, 2])
assert s1.issuperset(s2) == True, 'issuperset true'
assert s2.issuperset(s1) == False, 'issuperset false'

# === Isdisjoint ===
s1 = set([1, 2])
s2 = set([3, 4])
s3 = set([2, 3])
assert s1.isdisjoint(s2) == True, 'isdisjoint true'
assert s1.isdisjoint(s3) == False, 'isdisjoint false'

# === Bool ===
assert bool(set()) == False, 'empty set is falsy'
assert bool(set([1])) == True, 'non-empty set is truthy'

# === repr ===
assert repr(set()) == 'set()', 'empty set repr'

# === Set literals ===
s = {1, 2, 3}
assert len(s) == 3, 'set literal len'

s = {1, 1, 2, 2, 3}
assert len(s) == 3, 'set literal deduplication'

# Set literal with expressions
x = 5
s = {x, x + 1, x + 2}
assert len(s) == 3, 'set literal with expressions'



from inline_snapshot import snapshot

import ouros


def test_simple_expression():
    m = ouros.Sandbox('1 + 2')
    assert m.run() == snapshot(3)


def test_arithmetic():
    m = ouros.Sandbox('10 * 5 - 3')
    assert m.run() == snapshot(47)


def test_string_concatenation():
    m = ouros.Sandbox('"hello" + " " + "world"')
    assert m.run() == snapshot('hello world')


def test_multiple_runs_same_instance():
    m = ouros.Sandbox('x * 2', inputs=['x'])
    assert m.run(inputs={'x': 5}) == snapshot(10)
    assert m.run(inputs={'x': 10}) == snapshot(20)
    assert m.run(inputs={'x': -3}) == snapshot(-6)


def test_repr_no_inputs():
    m = ouros.Sandbox('1 + 1')
    assert repr(m) == snapshot("Sandbox(<1 line of code>, script_name='main.py')")


def test_repr_with_inputs():
    m = ouros.Sandbox('x', inputs=['x', 'y'])
    assert repr(m) == snapshot('Sandbox(<1 line of code>, script_name=\'main.py\', inputs=["x", "y"])')


def test_repr_with_external_functions():
    m = ouros.Sandbox('foo()', external_functions=['foo'])
    assert repr(m) == snapshot('Sandbox(<1 line of code>, script_name=\'main.py\', external_functions=["foo"])')


def test_repr_with_inputs_and_external_functions():
    m = ouros.Sandbox('foo(x)', inputs=['x'], external_functions=['foo'])
    assert repr(m) == snapshot(
        'Sandbox(<1 line of code>, script_name=\'main.py\', inputs=["x"], external_functions=["foo"])'
    )


def test_multiline_code():
    code = """
x = 1
y = 2
x + y
"""
    m = ouros.Sandbox(code)
    assert m.run() == snapshot(3)


def test_function_definition_and_call():
    code = """
def add(a, b):
    return a + b

add(3, 4)
"""
    m = ouros.Sandbox(code)
    assert m.run() == snapshot(7)



from abc import ABC, abstractmethod
from typing import Callable, Protocol, TypeAlias, TypeVar, final, overload, runtime_checkable

from _typeshed import AnyStr_co, structseq

_T = TypeVar('_T')
environ: dict[str, str]

@overload
def getenv(key: str) -> str | None: ...
@overload
def getenv(key: str, default: _T) -> str | _T: ...
@final
class stat_result(structseq[float], tuple[int, int, int, int, int, int, int, float, float, float]):
    # The constructor of this class takes an iterable of variable length (though it must be at least 10).
    #
    # However, this class behaves like a tuple of 10 elements,
    # no matter how long the iterable supplied to the constructor is.
    # https://github.com/python/typeshed/pull/6560#discussion_r767162532
    #
    # The 10 elements always present are st_mode, st_ino, st_dev, st_nlink,
    # st_uid, st_gid, st_size, st_atime, st_mtime, st_ctime.
    #
    # More items may be added at the end by some implementations.

    @property
    def st_mode(self) -> int:
        """protection bits"""
        ...

    @property
    def st_ino(self) -> int:
        """inode"""
        ...

    @property
    def st_dev(self) -> int:
        """device"""
        ...

    @property
    def st_nlink(self) -> int:
        """number of hard links"""
        ...

    @property
    def st_uid(self) -> int:
        """user ID of owner"""
        ...

    @property
    def st_gid(self) -> int:
        """group ID of owner"""
        ...

    @property
    def st_size(self) -> int:
        """total size, in bytes"""
        ...

    @property
    def st_atime(self) -> float:
        """time of last access"""
        ...

    @property
    def st_mtime(self) -> float:
        """time of last modification"""
        ...

    @property
    def st_ctime(self) -> float:
        """time of last change"""
        ...

# (Samuel) PathLike is included here because it's used by pathlib

# mypy and pyright object to this being both ABC and Protocol.
# At runtime it inherits from ABC and is not a Protocol, but it will be
# on the allowlist for use as a Protocol starting in 3.14.
@runtime_checkable
class PathLike(ABC, Protocol[AnyStr_co]):  # type: ignore[misc]  # pyright: ignore[reportGeneralTypeIssues]
    __slots__ = ()
    @abstractmethod
    def __fspath__(self) -> AnyStr_co: ...

_Opener: TypeAlias = Callable[[str, int], int]



import bisect

# === bisect_left ===
a = [1, 3, 5, 7, 9]
assert bisect.bisect_left(a, 5) == 2, 'bisect_left existing'
assert bisect.bisect_left(a, 4) == 2, 'bisect_left between'
assert bisect.bisect_left(a, 0) == 0, 'bisect_left before all'
assert bisect.bisect_left(a, 10) == 5, 'bisect_left after all'

# === bisect_right ===
assert bisect.bisect_right(a, 5) == 3, 'bisect_right existing'
assert bisect.bisect_right(a, 4) == 2, 'bisect_right between'
assert bisect.bisect_right(a, 0) == 0, 'bisect_right before all'
assert bisect.bisect_right(a, 10) == 5, 'bisect_right after all'

# === bisect is alias for bisect_right ===
assert bisect.bisect(a, 5) == 3, 'bisect alias'

# === bisect_left with duplicates ===
b = [1, 3, 3, 3, 5]
assert bisect.bisect_left(b, 3) == 1, 'bisect_left duplicates'
assert bisect.bisect_right(b, 3) == 4, 'bisect_right duplicates'

# === insort_left ===
c = [1, 3, 5, 7]
bisect.insort_left(c, 4)
assert c == [1, 3, 4, 5, 7], 'insort_left'

# === insort_right ===
d = [1, 3, 5, 7]
bisect.insort_right(d, 5)
assert d == [1, 3, 5, 5, 7], 'insort_right'

# === insort is alias for insort_right ===
e = [1, 3, 5]
bisect.insort(e, 3)
assert e == [1, 3, 3, 5], 'insort alias'

# === empty list ===
f = []
assert bisect.bisect_left(f, 1) == 0, 'bisect_left empty'
bisect.insort(f, 5)
assert f == [5], 'insort into empty'

# === lo/hi parameters for bisect_left ===
g = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
assert bisect.bisect_left(g, 5, 3, 8) == 4, 'bisect_left with lo and hi'
assert bisect.bisect_left(g, 5, 5) == 5, 'bisect_left with lo only'
assert bisect.bisect_left(g, 1, 0, 5) == 0, 'bisect_left lo=0 hi=5'

# === lo/hi parameters for bisect_right ===
assert bisect.bisect_right(g, 5, 3, 8) == 5, 'bisect_right with lo and hi'
assert bisect.bisect_right(g, 5, 5) == 5, 'bisect_right with lo only'

# === lo/hi parameters for insort_left ===
h = [1, 3, 5, 7, 9]
bisect.insort_left(h, 4, 1, 3)
assert h == [1, 3, 4, 5, 7, 9], 'insort_left with lo and hi'

# === lo/hi parameters for insort_right ===
j = [1, 3, 5, 7, 9]
bisect.insort_right(j, 6, 2, 4)
assert j == [1, 3, 5, 6, 7, 9], 'insort_right with lo and hi'

# === lo parameter only ===
k = [10, 20, 30, 40, 50]
assert bisect.bisect_left(k, 25, 2) == 2, 'bisect_left lo=2 searches from index 2'
assert bisect.bisect_right(k, 30, 2) == 3, 'bisect_right lo=2'

# === from import ===
from bisect import bisect_left, insort

m = [10, 20, 30]
assert bisect_left(m, 20) == 1, 'from import bisect_left'
insort(m, 25)
assert m == [10, 20, 25, 30], 'from import insort'

# === key functions ===
data = ['apple', 'banana', 'cherry']
assert bisect.bisect_left(data, 'banana', key=lambda x: x.lower()) == 1, 'bisect_left lambda key'
assert bisect.bisect_right(data, 'banana', key=lambda x: x.lower()) == 2, 'bisect_right lambda key'



# === dict() constructor with non-dict/non-iterable (lines 441-443, 449-451) ===
try:
    dict(42)
    assert False, 'dict(int) should raise TypeError'
except TypeError:
    pass

try:
    dict(True)
    assert False, 'dict(bool) should raise TypeError'
except TypeError:
    pass

# === dict() constructor copies dict (lines 463-473) ===
original = {'a': 1, 'b': 2}
copy = dict(original)
assert copy == {'a': 1, 'b': 2}, 'dict() copies dict'
assert copy is not original, 'dict() creates new dict'
original['c'] = 3
assert 'c' not in copy, 'dict() copy is independent'

# === Dict equality with different lengths (line 555) ===
assert ({'a': 1} == {'a': 1, 'b': 2}) == False, 'eq different length false'
assert ({'a': 1, 'b': 2} == {'a': 1}) == False, 'eq different length false reverse'

# === Dict equality with different values (line 566) ===
assert ({'a': 1, 'b': 2} == {'a': 1, 'b': 3}) == False, 'eq same keys different values'

# === Dict bool (line 610) ===
assert bool({}) == False, 'empty dict is falsy'
assert bool({'a': 1}) == True, 'non-empty dict is truthy'

# === Dict attribute error (line 656) ===
try:
    d = {'a': 1}
    d.nonexistent()
    assert False, 'unknown dict method should raise AttributeError'
except AttributeError as e:
    assert 'dict' in str(e), 'error mentions dict'

# === Dict.pop() missing key without default raises KeyError (line 727) ===
d = {'a': 1}
try:
    d.pop('missing')
    assert False, 'pop missing key no default should raise KeyError'
except KeyError:
    pass

# === Unknown dict method on class (line 759) ===
try:
    d = {'a': 1}
    d.bogus_method()
    assert False, 'bogus method should raise AttributeError'
except AttributeError:
    pass

# === dict.update() too many positional args (lines 810-818) ===
try:
    d = {}
    d.update({'a': 1}, {'b': 2})
    assert False, 'update with 2 positional args should raise TypeError'
except TypeError:
    pass

# === dict.update() with iterable of pairs (lines 834-979) ===
# List of tuples
d = {}
d.update([('a', 1), ('b', 2)])
assert d == {'a': 1, 'b': 2}, 'update with list of tuples'

# Tuple of tuples
d = {}
d.update((('x', 10), ('y', 20)))
assert d == {'x': 10, 'y': 20}, 'update with tuple of tuples'

# Update with dict (fast path, lines 834-863)
d = {'a': 1}
other = {'b': 2, 'c': 3}
d.update(other)
assert d == {'a': 1, 'b': 2, 'c': 3}, 'update with dict fast path'

# Update overwrites existing keys
d = {'a': 1, 'b': 2}
d.update({'b': 20, 'c': 30})
assert d == {'a': 1, 'b': 20, 'c': 30}, 'update dict overwrites'

# === dict.update() sequence element too short (lines 902-907, 920-927) ===
try:
    d = {}
    d.update([(), ('b', 2)])
    assert False, 'update with empty tuple should raise'
except (TypeError, ValueError) as e:
    assert 'length' in str(e).lower(), 'error mentions length'

try:
    d = {}
    d.update([('a',)])
    assert False, 'update with 1-element tuple should raise'
except (TypeError, ValueError) as e:
    assert 'length' in str(e).lower(), 'error mentions length for 1-element'

# === dict.update() sequence element too long (lines 940-957) ===
try:
    d = {}
    d.update([('a', 1, 'extra')])
    assert False, 'update with 3-element tuple should raise'
except (TypeError, ValueError) as e:
    assert 'length' in str(e).lower(), 'error mentions length for 3-element'

# === dict.setdefault() with unhashable key (lines 1031-1034) ===
try:
    d = {'a': 1}
    d.setdefault([1, 2], 'default')
    assert False, 'setdefault with list key should raise TypeError'
except TypeError:
    pass

# === dict.setdefault() returns existing and inserts new ===
d = {'a': 1}
assert d.setdefault('a', 99) == 1, 'setdefault existing returns value'
assert d.setdefault('b') is None, 'setdefault new returns None'
assert d['b'] is None, 'setdefault inserts None'
assert d.setdefault('c', 42) == 42, 'setdefault new returns default'
assert d['c'] == 42, 'setdefault inserts default'

# === dict.popitem() on empty dict (lines 1048-1049 via popitem_empty test, but exercise here too) ===
try:
    d = {}
    d.popitem()
    assert False, 'popitem on empty dict should raise KeyError'
except KeyError:
    pass

# === dict.fromkeys() with tuple iterable (lines 1132-1134) ===
d = dict.fromkeys((1, 2, 3), 'val')
assert d == {1: 'val', 2: 'val', 3: 'val'}, 'fromkeys with tuple iterable'

# === dict.fromkeys() with non-iterable (lines 1132-1134) ===
try:
    dict.fromkeys(42)
    assert False, 'fromkeys with int should raise TypeError'
except TypeError:
    pass

# === Dict comprehension ===
d = {k: k * 2 for k in range(3)}
assert d == {0: 0, 1: 2, 2: 4}, 'dict comprehension'

# === Dict in operator ===
d = {'a': 1, 'b': 2}
assert 'a' in d, 'key in dict'
assert 'c' not in d, 'key not in dict'
assert 1 not in d, 'value not in dict (checks keys)'

# === Dict iteration ===
d = {'a': 1, 'b': 2, 'c': 3}
keys = []
for k in d:
    keys.append(k)
assert keys == ['a', 'b', 'c'], 'dict iteration yields keys in order'

# === Dict keys/values/items (already tested but more coverage) ===
d = {'x': 10, 'y': 20}
assert list(d.keys()) == ['x', 'y'], 'keys returns list of keys'
assert list(d.values()) == [10, 20], 'values returns list of values'
assert list(d.items()) == [('x', 10), ('y', 20)], 'items returns list of tuples'

# === Dict merge operator | (if supported) ===
d1 = {'a': 1}
d2 = {'b': 2}
try:
    merged = d1 | d2
    assert merged == {'a': 1, 'b': 2}, 'dict merge operator'
except TypeError:
    pass  # operator may not be implemented yet

# === Dict with bool keys ===
d = {True: 'yes', False: 'no'}
assert d[True] == 'yes', 'bool key True'
assert d[False] == 'no', 'bool key False'

# === Dict.get() with unhashable key (lines 262-268) ===
try:
    d = {'a': 1}
    d.get([1, 2])
    assert False, 'get with list key should raise TypeError'
except TypeError:
    pass

# === Empty dict repr ===
assert repr({}) == '{}', 'empty dict repr'

# === Dict with nested values ===
d = {'a': [1, 2], 'b': {'c': 3}}
assert d['a'] == [1, 2], 'nested list value'
assert d['b'] == {'c': 3}, 'nested dict value'

# === Dict.copy() is shallow ===
inner = [1, 2, 3]
d = {'a': inner}
d2 = d.copy()
assert d2['a'] is inner, 'copy is shallow - same inner object'

# === Dict.clear() on dict with refs ===
d = {'a': [1, 2], 'b': 'hello'}
d.clear()
assert d == {}, 'clear removes all including ref values'
assert len(d) == 0, 'clear results in len 0'

# === Dict.update() with kwargs only ===
d = {}
d.update(x=1, y=2, z=3)
assert d == {'x': 1, 'y': 2, 'z': 3}, 'update kwargs only'

# === Dict.update() with both dict and kwargs ===
d = {}
d.update({'a': 1}, b=2)
assert d == {'a': 1, 'b': 2}, 'update dict + kwargs'

# === Dict deletion ===
d = {'a': 1, 'b': 2, 'c': 3}
del d['b']
assert d == {'a': 1, 'c': 3}, 'del removes key'
assert len(d) == 2, 'del decrements length'

try:
    del d['missing']
    assert False, 'del missing key should raise KeyError'
except KeyError:
    pass

# === Dict with None key ===
d = {None: 'nothing'}
assert d[None] == 'nothing', 'None as dict key'

# === Dict with tuple keys ===
d = {(1, 2): 'pair'}
assert d[(1, 2)] == 'pair', 'tuple key lookup'



import codecs
import sys


# === Public API Surface ===
public_names = [
    'BOM',
    'BOM32_BE',
    'BOM32_LE',
    'BOM64_BE',
    'BOM64_LE',
    'BOM_BE',
    'BOM_LE',
    'BOM_UTF16',
    'BOM_UTF16_BE',
    'BOM_UTF16_LE',
    'BOM_UTF32',
    'BOM_UTF32_BE',
    'BOM_UTF32_LE',
    'BOM_UTF8',
    'BufferedIncrementalDecoder',
    'BufferedIncrementalEncoder',
    'Codec',
    'CodecInfo',
    'EncodedFile',
    'IncrementalDecoder',
    'IncrementalEncoder',
    'StreamReader',
    'StreamReaderWriter',
    'StreamRecoder',
    'StreamWriter',
    'ascii_decode',
    'ascii_encode',
    'backslashreplace_errors',
    'charmap_build',
    'charmap_decode',
    'charmap_encode',
    'decode',
    'encode',
    'escape_decode',
    'escape_encode',
    'getdecoder',
    'getencoder',
    'getincrementaldecoder',
    'getincrementalencoder',
    'getreader',
    'getwriter',
    'ignore_errors',
    'iterdecode',
    'iterencode',
    'latin_1_decode',
    'latin_1_encode',
    'lookup',
    'lookup_error',
    'make_encoding_map',
    'make_identity_dict',
    'namereplace_errors',
    'open',
    'raw_unicode_escape_decode',
    'raw_unicode_escape_encode',
    'readbuffer_encode',
    'register',
    'register_error',
    'replace_errors',
    'strict_errors',
    'unicode_escape_decode',
    'unicode_escape_encode',
    'unregister',
    'utf_16_be_decode',
    'utf_16_be_encode',
    'utf_16_decode',
    'utf_16_encode',
    'utf_16_ex_decode',
    'utf_16_le_decode',
    'utf_16_le_encode',
    'utf_32_be_decode',
    'utf_32_be_encode',
    'utf_32_decode',
    'utf_32_encode',
    'utf_32_ex_decode',
    'utf_32_le_decode',
    'utf_32_le_encode',
    'utf_7_decode',
    'utf_7_encode',
    'utf_8_decode',
    'utf_8_encode',
    'xmlcharrefreplace_errors',
]

for name in public_names:
    assert hasattr(codecs, name), f'missing codecs.{name}'


# === BOM constants ===
assert codecs.BOM_BE == b'\xfe\xff', 'BOM_BE'
assert codecs.BOM_LE == b'\xff\xfe', 'BOM_LE'
assert codecs.BOM_UTF16_BE == b'\xfe\xff', 'BOM_UTF16_BE'
assert codecs.BOM_UTF16_LE == b'\xff\xfe', 'BOM_UTF16_LE'
assert codecs.BOM_UTF32_BE == b'\x00\x00\xfe\xff', 'BOM_UTF32_BE'
assert codecs.BOM_UTF32_LE == b'\xff\xfe\x00\x00', 'BOM_UTF32_LE'
assert codecs.BOM_UTF8 == b'\xef\xbb\xbf', 'BOM_UTF8'
assert codecs.BOM32_BE == b'\xfe\xff', 'BOM32_BE'
assert codecs.BOM32_LE == b'\xff\xfe', 'BOM32_LE'
assert codecs.BOM64_BE == b'\x00\x00\xfe\xff', 'BOM64_BE'
assert codecs.BOM64_LE == b'\xff\xfe\x00\x00', 'BOM64_LE'

if sys.byteorder == 'little':
    assert codecs.BOM == codecs.BOM_LE, 'native BOM on little-endian'
    assert codecs.BOM_UTF16 == codecs.BOM_UTF16_LE, 'native UTF16 BOM on little-endian'
    assert codecs.BOM_UTF32 == codecs.BOM_UTF32_LE, 'native UTF32 BOM on little-endian'
else:
    assert codecs.BOM == codecs.BOM_BE, 'native BOM on big-endian'
    assert codecs.BOM_UTF16 == codecs.BOM_UTF16_BE, 'native UTF16 BOM on big-endian'
    assert codecs.BOM_UTF32 == codecs.BOM_UTF32_BE, 'native UTF32 BOM on big-endian'


# === Top-level encode/decode ===
assert codecs.encode('abc') == b'abc', 'codecs.encode utf-8'
assert codecs.decode(b'abc') == 'abc', 'codecs.decode utf-8'
assert codecs.encode('é', 'latin-1') == b'\xe9', 'codecs.encode latin-1'
assert codecs.decode(b'\xe9', 'latin-1') == 'é', 'codecs.decode latin-1'


# === Lookup and helper getters ===
info = codecs.lookup('utf-8')
assert len(info) == 4, 'lookup tuple-like length'

enc = codecs.getencoder('utf-8')
dec = codecs.getdecoder('utf-8')
assert enc('a')[0] == b'a', 'getencoder returns callable'
assert dec(b'a')[0] == 'a', 'getdecoder returns callable'
assert callable(codecs.getincrementalencoder('utf-8')), 'getincrementalencoder callable'
assert callable(codecs.getincrementaldecoder('utf-8')), 'getincrementaldecoder callable'
assert callable(codecs.getreader('utf-8')), 'getreader callable'
assert callable(codecs.getwriter('utf-8')), 'getwriter callable'


# === Iter helpers ===
assert list(codecs.iterencode(['a', 'b'], 'utf-8')) == [b'a', b'b'], 'iterencode'
assert list(codecs.iterdecode([b'a', b'b'], 'utf-8')) == ['a', 'b'], 'iterdecode'


# === ASCII / Latin-1 ===
assert codecs.ascii_encode('abc') == (b'abc', 3), 'ascii_encode'
assert codecs.ascii_decode(b'abc') == ('abc', 3), 'ascii_decode'
assert codecs.latin_1_encode('Aÿ') == (b'A\xff', 2), 'latin_1_encode'
assert codecs.latin_1_decode(b'\xff') == ('ÿ', 1), 'latin_1_decode'


# === UTF-8 / UTF-7 ===
assert codecs.utf_8_encode('☃') == (b'\xe2\x98\x83', 1), 'utf_8_encode'
assert codecs.utf_8_decode(b'\xe2\x98\x83') == ('☃', 3), 'utf_8_decode'
assert codecs.utf_7_encode('A+') == (b'A+-', 2), 'utf_7_encode'
assert codecs.utf_7_decode(b'A+-') == ('A+', 3), 'utf_7_decode'


# === UTF-16 family ===
utf16 = codecs.utf_16_encode('A')
assert utf16[1] == 1, 'utf_16_encode consumed chars'
assert codecs.utf_16_decode(utf16[0])[0] == 'A', 'utf_16_decode roundtrip'
utf16_ex = codecs.utf_16_ex_decode(utf16[0])
assert utf16_ex[0] == 'A', 'utf_16_ex_decode text'
assert utf16_ex[1] == len(utf16[0]), 'utf_16_ex_decode consumed bytes'
assert utf16_ex[2] in (-1, 1), 'utf_16_ex_decode byteorder'
assert codecs.utf_16_le_decode(codecs.utf_16_le_encode('A')[0])[0] == 'A', 'utf_16_le roundtrip'
assert codecs.utf_16_be_decode(codecs.utf_16_be_encode('A')[0])[0] == 'A', 'utf_16_be roundtrip'


# === UTF-32 family ===
utf32 = codecs.utf_32_encode('A')
assert utf32[1] == 1, 'utf_32_encode consumed chars'
assert codecs.utf_32_decode(utf32[0])[0] == 'A', 'utf_32_decode roundtrip'
utf32_ex = codecs.utf_32_ex_decode(utf32[0])
assert utf32_ex[0] == 'A', 'utf_32_ex_decode text'
assert utf32_ex[1] == len(utf32[0]), 'utf_32_ex_decode consumed bytes'
assert utf32_ex[2] in (-1, 1), 'utf_32_ex_decode byteorder'
assert codecs.utf_32_le_decode(codecs.utf_32_le_encode('A')[0])[0] == 'A', 'utf_32_le roundtrip'
assert codecs.utf_32_be_decode(codecs.utf_32_be_encode('A')[0])[0] == 'A', 'utf_32_be roundtrip'


# === Escape codecs ===
assert codecs.unicode_escape_encode('A\n') == (b'A\\n', 2), 'unicode_escape_encode'
assert codecs.unicode_escape_decode(b'A\\n') == ('A\n', 3), 'unicode_escape_decode'
assert codecs.raw_unicode_escape_encode('A☃') == (b'A\\u2603', 2), 'raw_unicode_escape_encode'
assert codecs.raw_unicode_escape_decode(b'A\\u2603')[0] == 'A☃', 'raw_unicode_escape_decode'
assert codecs.escape_encode(b'A\n') == (b'A\\n', 2), 'escape_encode'
assert codecs.escape_decode(b'A\\n') == (b'A\n', 3), 'escape_decode'


# === Charmap helpers ===
assert codecs.charmap_encode('abc', None, {ord('a'): 1, ord('b'): 2, ord('c'): 3}) == (b'\x01\x02\x03', 3), 'charmap_encode'
assert codecs.charmap_decode(b'\x01\x02', None, {1: 'a', 2: 'b'}) == ('ab', 2), 'charmap_decode'
assert codecs.charmap_build('abc') == {97: 0, 98: 1, 99: 2}, 'charmap_build'

identity = codecs.make_identity_dict(range(3))
assert identity == {0: 0, 1: 1, 2: 2}, 'make_identity_dict'

encoding_map = codecs.make_encoding_map({1: 'a', 2: 'b'})
assert encoding_map == {'a': 1, 'b': 2}, 'make_encoding_map'


# === readbuffer_encode ===
assert codecs.readbuffer_encode(b'abc') == (b'abc', 3), 'readbuffer_encode bytes'
assert codecs.readbuffer_encode('abc') == (b'abc', 3), 'readbuffer_encode str'


# === Error registry helpers ===
codecs.register(lambda _name: None)
codecs.unregister(lambda _name: None)
codecs.register_error('ouros_handler', lambda exc: ('?', 1))
assert callable(codecs.lookup_error('strict')), 'lookup_error strict'
assert callable(codecs.lookup_error('ignore')), 'lookup_error ignore'
assert callable(codecs.lookup_error('replace')), 'lookup_error replace'
assert callable(codecs.lookup_error('backslashreplace')), 'lookup_error backslashreplace'
assert callable(codecs.lookup_error('xmlcharrefreplace')), 'lookup_error xmlcharrefreplace'
assert callable(codecs.lookup_error('namereplace')), 'lookup_error namereplace'


# === Error callbacks are present and callable ===
assert callable(codecs.strict_errors), 'strict_errors callable'
assert callable(codecs.ignore_errors), 'ignore_errors callable'
assert callable(codecs.replace_errors), 'replace_errors callable'
assert callable(codecs.backslashreplace_errors), 'backslashreplace_errors callable'
assert callable(codecs.xmlcharrefreplace_errors), 'xmlcharrefreplace_errors callable'
assert callable(codecs.namereplace_errors), 'namereplace_errors callable'


# === I/O APIs exist (sandbox-safe call behavior differs by runtime) ===
try:
    codecs.open('__ouros_nonexistent_file__', 'r')
except Exception as exc:
    assert exc is not None, 'codecs.open raises'

try:
    codecs.EncodedFile(object(), 'utf-8')
except Exception as exc:
    assert exc is not None, 'codecs.EncodedFile raises'


# === Module attrs mirrored from CPython ===
assert hasattr(codecs, 'builtins'), 'codecs.builtins'
assert hasattr(codecs, 'sys'), 'codecs.sys'
